import abc as _abc
import copy
import logging
import warnings as _warnings
from itertools import chain
from typing import TYPE_CHECKING
from typing import Any as _Any
from typing import Dict as _Dict
from typing import List as _List
from typing import Mapping as _Mapping
from typing import Tuple as _Tuple
from typing import Union as _Union

import torch.nn as _nn

from tamm import _adapters_v1
from tamm import _warnings as _tamm_warnings
from tamm.layers import common as layer_common
from tamm.model_repo._warning import _warn_deprecated_published_config
from tamm.model_repo.exceptions import UnrecognizedModelIdentifierError
from tamm.model_repo.publishing import PublishedModelConfig as _PublishedModelConfig

if TYPE_CHECKING:
    from tamm.models.common import ModelConfig
logger = logging.getLogger(__name__)


_LORA_MODEL_ADAPTER_TYPES: _Tuple[_Any, ...] = (_adapters_v1.LoRAModelAdapter,)
_MULTI_LORA_MODEL_ADAPTER_TYPES: _Tuple[_Any, ...] = (
    _adapters_v1.MultiLoRAModelAdapter,
)


def _get_adapters_from_config(config: "ModelConfig"):
    sub_modules = ["decoder", "image_tokenizer"]
    adapters = []
    for sub_module in sub_modules:
        if hasattr(config, sub_module):
            sub_module_adapters = _get_adapters_from_config(getattr(config, sub_module))
            if sub_module_adapters:
                adapters.extend(sub_module_adapters)
    if (
        hasattr(config, "adapters")
        and isinstance(config.adapters, dict)
        and len(config.adapters) > 0
    ):
        adapters.append(config.adapters)
    return adapters


def _get_lora_adapters(adapters):
    result = []
    if adapters is None:
        return result

    for adapter in adapters.values():
        if isinstance(adapter, _MULTI_LORA_MODEL_ADAPTER_TYPES):
            sub_adapters = _get_lora_adapters(adapter.lora_model_adapters)
            result.extend(sub_adapters)
        elif isinstance(adapter, _LORA_MODEL_ADAPTER_TYPES):
            result.append(adapter)

    return result


class BaseModelRepo(metaclass=_abc.ABCMeta):
    """
    Base class of all model repos, defines the common interface that must be implemented by concrete backends.
    """

    def __init__(self, *args, **kwargs):
        del args, kwargs

    @_abc.abstractmethod
    def _get_configs_impl(self) -> _Mapping[str, "_PublishedModelConfig"]:
        """
        Returns a mapping from model ids to PublishedModelConfig
        """
        raise NotImplementedError(
            "_get_configs_impl() must be implemented"
            " for concrete classes inherited from BaseModelRepo"
        )

    @_abc.abstractmethod
    def is_alive(self) -> bool:
        """
        Returns whether the model repo is ready to serve models
        """

    def _list_models(
        self,
        include_descriptions=False,
        filter_deprecated=False,
        exclude_adapted=True,
    ) -> _Union[_List[str], _Dict[str, str]]:
        """
        Returns a list of model ids that can be created by this repo
        """
        _configs = [
            config
            for config in self._get_configs(
                filter_deprecated=filter_deprecated
            ).values()
            if exclude_adapted ^ config.model_config.has_adapters
        ]

        if not include_descriptions:
            return [config.model_id for config in _configs]

        return {config.model_id: config.description for config in _configs}

    def list_models(
        self, include_descriptions: bool = False, filter_deprecated: bool = False
    ) -> _Union[_List[str], _Dict[str, str]]:
        """
        Returns a list of model ids that can be created by this repo
        """
        return self._list_models(
            include_descriptions=include_descriptions,
            filter_deprecated=filter_deprecated,
            exclude_adapted=True,
        )

    def list_adapted_models(
        self, include_descriptions=False, filter_deprecated=False
    ) -> _Union[_List[str], _Dict[str, str]]:
        return self._list_models(
            include_descriptions=include_descriptions,
            filter_deprecated=filter_deprecated,
            exclude_adapted=False,
        )

    def list_model_builders(self, include_descriptions=False, filter_deprecated=False):
        return self.list_model_configs(
            include_descriptions=include_descriptions,
            filter_deprecated=filter_deprecated,
        )

    def list_model_configs(
        self, include_descriptions=False, filter_deprecated=False
    ) -> _Union[_List[str], _Dict[str, str]]:
        """
        Returns a list of model ids that can be created by this repo
        """
        items = self._get_configs(filter_deprecated=filter_deprecated)
        if not include_descriptions:
            return [item.model_id for item in items.values()]
        return {item.model_id: item.description for item in items.values()}

    def list_available_lora_adapted_models(self) -> _Dict[str, _Dict[int, str]]:
        """
        Helper method to retrieve the mapping from base model to all
        available adapted model.
        """
        result: _Dict[str, _Dict[int, str]] = {}
        all_model_ids = sorted(set(self.list_adapted_models() + self.list_models()))  # type: ignore[operator]
        for model_id in all_model_ids:
            with _warnings.catch_warnings(record=True):
                config = self.create_model_config(model_id)
            backbone_id = config.metadata.source_model_details.get(
                "backbone_tamm_model_id", None
            )
            if backbone_id is None:
                continue
            adapters = _get_adapters_from_config(config)

            if adapters is None or len(adapters) != 1:
                continue
            adapter = list(adapters[0].values())[0]

            if not isinstance(adapter, _LORA_MODEL_ADAPTER_TYPES):
                continue
            group = result.setdefault(backbone_id, {})
            if adapter.rank in group:
                raise RuntimeError(
                    f"Multiple adapted models found for {backbone_id} rank {adapter.rank}"
                )
            group[adapter.rank] = model_id

        for backbone_id, group in result.items():
            result[backbone_id] = dict(sorted(group.items()))

        return result

    def create_adapted_model(self, model_name: str, *args, **kwargs) -> "_nn.Module":
        """
        Creates a new adapted model

        .. warning::

            :func:`create_adapted_model` is deprecated. Please use
            :func:`create_model` or :func:`create_model_config` instead.
        """
        _tamm_warnings.deprecation(
            "create_adapted_model() is deprecated.  Please use create_model() "
            "or create_model_config() instead."
        )
        # Separate LoRA options from **kwargs:
        lora_key_mapping = {
            "rank": "rank",
            "alpha": "alpha",
            "adapt_q": "adapt_attention_queries",
            "adapt_k": "adapt_attention_keys",
            "adapt_v": "adapt_attention_values",
            "lora_dropout_p": "dropout_p",
        }
        lora_options = {
            v: kwargs.pop(k) for k, v in lora_key_mapping.items() if k in kwargs
        }

        # Create config without LoRA options:
        model_config = self.create_model_config(model_name, **kwargs)

        # Update LoRA adapters:
        for adapter in _get_lora_adapters(model_config.adapters):
            for key, val in lora_options.items():
                setattr(adapter, key, val)

        return model_config.create_model(*args, **kwargs)

    def create_model_builder(
        self, model_name: str, *args, **kwargs
    ) -> "layer_common.LayerBuilder":
        """
        Creates a new model builder
        """
        config = self.create_model_config(model_name, *args, **kwargs)
        return config.create_builder(*args, **kwargs)

    def create_model(self, model_name: str, *args, **kwargs) -> "_nn.Module":
        # test-create basic model config solely from model_name, ignore kwargs to determine if this model is adapted
        config_no_override = self.create_model_config(model_name)
        if not (
            hasattr(config_no_override, "adapters")
            and isinstance(config_no_override.adapters, dict)
            and len(config_no_override.adapters) > 0
        ):
            config = self.create_model_config(model_name, **kwargs)
            return config.create_model(*args, **kwargs)
        return self.create_adapted_model(model_name, *args, **kwargs)

    def create_model_config(self, model_name: str, *args, **kwargs) -> "ModelConfig":
        try:
            published_config = self._get_configs_impl()[model_name]
        except KeyError as e:
            raise KeyError(
                f"Model '{model_name}' does not exist in '{self.__class__.__name__}'"
            ) from e
        _warn_deprecated_published_config(published_config)
        published_config = copy.deepcopy(published_config)
        published_config.model_config.update_configured_args(*args, **kwargs)
        return published_config.model_config

    def is_model_name(self, model_name: str) -> bool:
        return model_name in self.list_models()

    def is_adapted_model_name(self, model_name: str) -> bool:
        return model_name in self.list_adapted_models()

    def is_model_builder_name(self, model_builder_name: str) -> bool:
        return model_builder_name in self.list_model_builders()

    def is_model_config_name(self, model_config_name: str) -> bool:
        return model_config_name in self.list_model_configs()

    def _get_configs(
        self, filter_deprecated: bool = False
    ) -> _Dict[str, "_PublishedModelConfig"]:
        configs = self._get_configs_impl()
        if filter_deprecated:
            configs = {
                model_name: published_config
                for model_name, published_config in configs.items()
                if not published_config.is_deprecated
            }

        return dict(sorted(configs.items()))

    @_abc.abstractmethod
    def clear_cache(self):
        """
        Clear any local cache this model repo may have used
        """

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def __add__(self, other: "BaseModelRepo"):
        """
        Model configs defined in any model repo will supersede model configs of the
        same model_id in model repo to its left.

        Args:
            other: ModelRepo

        Returns: Composite ModelRepo

        """
        return CompositeModelRepo(self, other)


class PlaceholderModelRepo(BaseModelRepo):
    """
    Placeholder model repo, contains no model definitions
    """

    def _get_configs_impl(self) -> _Mapping[str, "_PublishedModelConfig"]:
        return {}

    def is_alive(self) -> bool:
        return True

    def clear_cache(self) -> None:
        """
        No-op
        """


class CompositeModelRepo(BaseModelRepo):
    """
    Composite model repo which keeps tracks of many model repo instances and act like a union of all
    """

    def __init__(self, *repo_instances: "BaseModelRepo"):
        super().__init__(*repo_instances)
        self._repo_instances = repo_instances
        self._configs: _Dict[str, "_PublishedModelConfig"] = {}

    def _chained_iterator(self, method, *args, **kwargs):
        """
        Call each repo_instance's ``method`` with *args, anw **kwargs
        """
        return_values = [
            getattr(repo_instance, method)(*args, **kwargs)
            for repo_instance in self._repo_instances
        ]
        if any(isinstance(return_value, dict) for return_value in return_values):
            return {
                k: v for return_value in return_values for k, v in return_value.items()
            }
        return list(chain(*return_values))

    def _first_successful(self, method, *args, **kwargs):
        """
        Call each repo_instance's ``method`` with *args, anw **kwargs, return the first successful call
        """
        exceptions = []
        for repo_instance in self._repo_instances:
            try:
                return getattr(repo_instance, method)(*args, **kwargs)
            except Exception as e:
                exceptions.append(e)
                logger.debug(
                    "%s raised exception when calling %s(%s): %s",
                    repo_instance,
                    method,
                    args[0],
                    str(e),
                )
        # Raise the first meaningful exception from composite repo, UnrecognizedModelIdentifierError is
        # not meaningful because it basically means a model repo doesn't understand a specific URI schema.
        # However, if all model repos raise UnrecognizedModelIdentifierError we will still surface it to the user.
        if all(isinstance(e, UnrecognizedModelIdentifierError) for e in exceptions):
            raise next(e for e in exceptions)
        raise next(
            e for e in exceptions if not isinstance(e, UnrecognizedModelIdentifierError)
        )

    def create_model(self, model_name: str, *args, **kwargs) -> "_nn.Module":
        return self._first_successful("create_model", model_name, *args, **kwargs)

    def create_model_builder(
        self, model_name: str, *args, **kwargs
    ) -> "layer_common.LayerBuilder":
        return self._first_successful(
            "create_model_builder", model_name, *args, **kwargs
        )

    def create_adapted_model(self, model_name: str, *args, **kwargs) -> "_nn.Module":
        return self._first_successful(
            "create_adapted_model", model_name, *args, **kwargs
        )

    def create_model_config(self, model_name: str, *args, **kwargs) -> "ModelConfig":
        return self._first_successful(
            "create_model_config", model_name, *args, **kwargs
        )

    def list_model_builders(self, *args, **kwargs):
        return self._chained_iterator("list_model_builders", *args, **kwargs)

    def list_model_configs(self, *args, **kwargs):
        return self._chained_iterator("list_model_configs", *args, **kwargs)

    def list_available_lora_adapted_models(self, *args, **kwargs):
        return self._chained_iterator(
            "list_available_lora_adapted_models", *args, **kwargs
        )

    def list_adapted_models(self, *args, **kwargs):
        return self._chained_iterator("list_adapted_models", *args, **kwargs)

    def list_models(self, *args, **kwargs):
        return self._chained_iterator("list_models", *args, **kwargs)

    def is_alive(self) -> bool:
        return all(repo.is_alive() for repo in self._repo_instances)

    def _get_configs_impl(self) -> _Dict[str, "_PublishedModelConfig"]:
        """
        When model repos are added for composition, for example,
        composite_model_repo = model_repo_a + model_repo_b

        model_repo_b's model config will supersede model_repo_a's model config if they
        share the same model_id. Any model repo in the addition will supersede
        model repo to its left.

        Returns: A dictionary mapping model_id->PublishedModelConfig instance

        """
        # pylint:disable=protected-access
        for repo in self._repo_instances:
            self._configs.update(repo._get_configs_impl())
        return self._configs

    def __repr__(self):
        repo_instances_str = ", ".join([str(repo) for repo in self._repo_instances])
        return f"{self.__class__.__name__}({repo_instances_str})"

    def clear_cache(self):
        for repo in self._repo_instances:
            repo.clear_cache()



from tamm.layers.transformer import token_metadata
from tamm.layers.transformer.attention_mask import AttentionMask
from tamm.layers.transformer.kv_cache import (
    BaseKVCache,
    BaseKVCacheView,
    KVCacheLayerView,
    VanillaKVCache,
    VanillaKVCacheView,
)
from tamm.layers.transformer.layer import TransformerLayer
from tamm.layers.transformer.layer_sequence import (
    BaseTransformerLayerSequence,
    KVReuseTransformerLayerSequence,
    TransformerLayerSequence,
    UniformTransformerLayerSequence,
)
from tamm.layers.transformer.positional_encoding import (
    ALiBiPositionalEncoding,
    NoPositionalEncoding,
    RotaryPositionalEmbedding,
    SecondaryPositionalEncodings,
    SequentialPositionalEncoding,
    SinusoidalPositionalEmbedding,
    SlidingWindowPositionalEncoding,
    TransformerAbsolutePositionalEmbedding,
    TransformerPositionalEncoding,
)
from tamm.layers.transformer.stack import ExtendedTransformerStack, TransformerStack
from tamm.layers.transformer.text import (
    CausalLMTransformer,
    CausalLMTransformerOutput,
    TextTransformerEncoder,
)
from tamm.layers.transformer.token_type_encoding import (
    TransformerTokenTypeEmbedding,
    TransformerTokenTypeEncoding,
)
from tamm.layers.transformer.vision import VisionTransformerEncoder



import warnings as _warnings

_TAMM_CONFIG_TYPE_TO_HF_MODEL_TYPE = {}


def _register_hf_model_type(*, tamm_config_type):
    """
    Associates a HF model type with a tamm config type so that we can create the HF
    model from the tamm config.
    """

    def decorator(hf_cls):
        if tamm_config_type in _TAMM_CONFIG_TYPE_TO_HF_MODEL_TYPE:
            _warnings.warn(f"Overwriting HF model type for {tamm_config_type}")

        _TAMM_CONFIG_TYPE_TO_HF_MODEL_TYPE[tamm_config_type] = hf_cls
        return hf_cls

    return decorator


def _get_hf_model_type(tamm_config_type):
    return _TAMM_CONFIG_TYPE_TO_HF_MODEL_TYPE.get(tamm_config_type)



"""
This module provides a class for wrapping arbitary functions into :obj:`nn.Module`
objects.
"""

from typing import Callable as _Callable

import torch.nn as _nn

from tamm.layers.common import LayerMixin as _LayerMixin


class Lambda(_nn.Module, _LayerMixin):
    """
    A :class:`nn.Module` for wrapping a function as a :obj:`Module`.

    Args:
        function (:obj:`Callable`): A callable function that takes inputs to the forward
            function and returns the output of forward.
    """

    def __init__(self, function: _Callable):
        super().__init__()
        self.function = function

    # pylint: disable-next=all
    def forward(self, *args, **kwargs):
        return self.function(*args, **kwargs)

    def extra_repr(self) -> str:
        return f"function={repr(self.function)}"



from tamm._plugin import utils as plugin_utils
from tamm.runtime_configuration import rc


def execute_core_import_callbacks() -> None:
    entrypoints = plugin_utils.discover_plugin_extras(
        rc.PLUGINS_EXTRAS_ENTRYPOINT_NAMES.core_import_callback
    )
    for entrypoint in entrypoints:
        plugin_utils.execute_plugin_object_reference(entrypoint.value, call_obj=True)



"""
layers.sequential
^^^^^^^^^^^^^^^^^

This module implements a generalization of ``torch.nn.Sequential``.  The |tamm| version
supports more complex sequential scenarios and includes features to help with training
optimizations such as FSDP.

.. autoclass:: tamm.layers.Sequential
    :show-inheritance:
    :members:
"""

import collections as _collections
import dataclasses as _dataclasses
import logging as _logging
from typing import Any as _Any
from typing import Dict as _Dict
from typing import List as _List
from typing import Optional as _Optional
from typing import Tuple as _Tuple
from typing import Union as _Union

import torch.nn as _nn
from torch.utils import checkpoint as _torch_utils_checkpoint

from tamm import _helpers, _warnings
from tamm.layers import side_outputs as _side_outputs
from tamm.layers.common import LayerMixin as _LayerMixin
from tamm.typing import OptionalModuleOrBuilder as _OptionalModuleOrBuilder

_logger = _logging.getLogger(__name__)


@_dataclasses.dataclass
class _ActivationCheckpointingOptions:
    is_enabled: bool = False
    use_reentrant: bool = False
    preserve_rng_state: bool = True


class Sequential(_nn.Module, _LayerMixin):
    """
    This layer generalizes ``torch.nn.Sequential``.  Like the native PyTorch version,
    its output is the chained output of child layers.  Additional features include:

    * Side inputs: Rather than receiving as input only the prior layer's output, each
        layer has the option of receiving additional arguments "from the side."  Users
        pass side inputs to :meth:`Sequential.forward`, which forwards the args to the
        appropriate layers in the sequence.
    * Side outputs: Rather than returning only the output of the final layer, the
        forward pass can return additional outputs from intermediate layers.
    * Groupable layers: The :meth:`group_subsequences_by_lengths` method reorganizes the
        ``nn.Module`` graph by replacing contiguous subsequences of layers with one
        :class:`Sequential` module per subsequence.  This provides flexibility to apply
        training optimizations to arbitrary subsequences of layers.  This reorganization
        does not affect the output of the forward pass.

    This layer implements the full interface of ``torch.nn.Sequential``, which means
    that code that supports ``torch.nn.Sequential`` should also support this layer.

    Args:
        named_layers (:obj:`dict` that maps :obj:`str` to :obj:`nn.Module`): A mapping
            of layer names to layer objects.  This defines the layer sequence, including
            its ordering (the sequence follows the same ordering as the :obj:`dict`).
        residual_connection (:obj:`_ModuleOrBuilder`): A Residual Connection layer,
            which when provided, applies that residual connection to the output of the
            :class: `Sequential` instance.
        side_input_keys (:obj:`dict` that maps :obj:`str` to a list of either \
            :obj:`str` or pairs of :obj:`str`, optional): A mapping of layer names to
            side argument names.  A value of ``["a", "c"]`` for ``layer_1`` means that
            the :class:`Sequential` instance forwards keyword arguments ``a`` and ``c``
            to ``layer_1`` during forward computation.  A value of
            ``["a", ("c", "new_c")]`` for ``layer_1`` means that the layer forwards
            ``a`` and ``c`` but uses the keyword ``new_c`` when passing the ``c``
            argument to ``layer_1``.  A value of ``["a", ("c", "**c")]`` means that
            the layer forwards ``a`` and ``c`` but passes ``c`` as ``**c``
            (``c`` is a dictionary of keyword arguments, which the layer unpacks when
            calling ``layer_1``).
        side_output_keys (:obj:`dict` that maps :obj:`str` to :obj:`str`, optional): A
            mapping of layer names to output names.  For example, a value of
            ``"hidden_state"`` for ``hidden_layer`` configures the forward pass to also
            return the output of ``hidden_layer`` as a side output with key
            ``"hidden_state"``.
        unpack_tuple_inputs (:obj:`bool`, optional): A flag that when ``True``,
            configures the forward pass to unpack inputs when calling each layer if the
            input is a :obj:`tuple`.  This enables a layer to receive multiple
            positional inputs if the output of the preceding layer is a :obj:`tuple`.

    """

    def __init__(
        self,
        named_layers: _Dict[str, _nn.Module],
        *,
        residual_connection: _OptionalModuleOrBuilder = None,
        side_input_keys: _Optional[
            _Dict[str, _List[_Union[str, _Tuple[str, str]]]]
        ] = None,
        side_output_keys: _Optional[_Dict[str, str]] = None,
        unpack_tuple_inputs: bool = False,
    ):
        _nn.Module.__init__(self)
        self._init_side_input_keys(side_input_keys)
        self._side_output_keys = (
            side_output_keys if side_output_keys is not None else {}
        )

        for layer_name in list(named_layers):
            if named_layers[layer_name] is not None:
                continue
            setattr(self, layer_name, None)
            named_layers.pop(layer_name)

        if residual_connection is not None:
            self._side_input_keys["residual_connection"].append(
                ("residual_input", "residual_input")
            )
            named_layers["residual_connection"] = residual_connection

        self._unpack_tuple_inputs = unpack_tuple_inputs
        for name, layer in named_layers.items():
            layer = _helpers.maybe_build_module(layer)
            self.add_module(name, layer)

        if "residual_connection" not in self.layer_names:
            self.residual_connection = None

        self._activation_checkpointing_options = _ActivationCheckpointingOptions()

    def _init_side_input_keys(self, side_input_keys):
        self._side_input_keys = _collections.defaultdict(list)
        if side_input_keys is None:
            side_input_keys = {}
        for layer_name, input_names in side_input_keys.items():
            for input_name in input_names:
                _logger.debug("init side input key %s", input_name)
                if isinstance(input_name, str):
                    input_name = (input_name, input_name)
                else:
                    input_name = tuple(input_name)
                    if len(input_name) != 2:
                        raise ValueError(
                            f"side input key does not have length 2: {input_name}"
                        )
                self._side_input_keys[layer_name].append(input_name)

    def __repr__(self):
        # Fall back if there is no more than 1 layers
        if len(self.named_layers) <= 1:
            return super().__repr__()

        class _REPRItem:
            def __init__(self, layer, name, index):
                self.layer = layer
                self.layer_repr = repr(layer)
                self.name = name
                self.index = index
                self.num_duplicated_layers = 1

            def __str__(self):
                lines = self.layer_repr.split("\n")
                if self.num_duplicated_layers > 1:
                    lines[0] = (
                        f"({self.index} - {self.index + self.num_duplicated_layers - 1}):"
                        f" {self.num_duplicated_layers} x " + lines[0]
                    )
                else:
                    lines[0] = f"({self.name}): " + lines[0]
                lines = [f"  {line}" for line in lines]
                return "\n".join(lines)

        all_repr_items = [
            _REPRItem(layer, name, index)
            for index, (name, layer) in enumerate(self.named_layers)
        ]

        idx = 0
        while idx < len(all_repr_items) - 1:
            if all_repr_items[idx].layer_repr == all_repr_items[idx + 1].layer_repr:
                all_repr_items.pop(idx + 1)
                all_repr_items[idx].num_duplicated_layers += 1
            else:
                idx += 1

        extra_repr = self.extra_repr()
        if extra_repr:
            extra_repr = f"\n  {extra_repr}"

        return (
            f"{self.__class__.__name__}(\n"
            + "\n".join(str(_) for _ in all_repr_items)
            + extra_repr
            + "\n)"
        )

    @property
    def side_input_keys(self) -> _Dict[str, _List[str]]:
        """The side inputs configuration."""
        return self._side_input_keys

    @property
    def has_residual_connection(self) -> bool:
        """Whether the layer has a residual connection."""
        return self.residual_connection is not None

    @property
    def side_output_keys(self) -> _Dict[str, str]:
        """The side outputs configuration."""
        return self._side_output_keys

    @property
    def unpack_tuple_inputs(self) -> bool:
        """The unpack_tuple_inputs setting."""
        return self._unpack_tuple_inputs

    @property
    def layer_names(self) -> _collections.abc.KeysView:
        """The names of layers in the sequence as a :obj:`KeysView`."""
        return self._modules.keys()

    @property
    def layers(self) -> _collections.abc.ValuesView:
        """The layers in the sequence as a :obj:`ValuesView`."""
        return self._modules.values()

    @property
    def named_layers(self) -> _collections.abc.ItemsView:
        """The names and layers in the sequence as an :obj:`ItemsView`."""
        return self._modules.items()

    @classmethod
    def from_nn_sequential(cls, module: _nn.Sequential) -> "Sequential":
        """
        Creates a new :obj:`Sequential` instance using the layer sequence from a
        :obj:`torch.nn.Sequential` module.
        """
        named_layers = {f"layer_{idx}": layer for idx, layer in enumerate(module)}
        return cls(named_layers)

    def register_side_input(
        self,
        layer_name: str,
        parent_input_name: str,
        child_input_name: _Optional[str] = None,
    ):
        """
        Register a side input argument for :meth:`forward`.

        Args:
            layer_name (:obj:`str`): The name of the layer that receives the side input
                during :meth:`forward`.
            parent_input_name (:obj:`str`): The name of the side input when passed to
                :meth:`Sequential.forward`.
            child_input_name (:obj:`str`, optional): The name for
                :meth:`Sequential.forward` to use when passing the side input to the
                child layer as a keyword argument.  This defaults to the
                ``parent_input_name`` value.
        """
        if child_input_name is None:
            child_input_name = parent_input_name
        input_pair = (parent_input_name, child_input_name)

        input_keys_for_layer = self._side_input_keys[layer_name]
        if input_pair not in input_keys_for_layer:
            input_keys_for_layer.append(input_pair)

    def unregister_side_input(
        self,
        layer_name: str,
        parent_input_name: str,
        child_input_name: str,
    ):
        """
        Un-register a side input argument.  This raises a :obj:`ValueError` if there is
        no side input registered with ``layer_name`` and ``parent_input_name``.

        Args:
            layer_name (:obj:`str`): The name of the layer that receives the side input
                during :meth:`forward`.
            parent_input_name (:obj:`str`): The name of the side input when passed to
                :meth:`Sequential.forward`.
            child_input_name (:obj:`str`, optional): The name for
                :meth:`Sequential.forward` to use when passing the side input to the
                child layer as a keyword argument.
        """
        input_pair = (parent_input_name, child_input_name)

        input_keys_for_layer = self._side_input_keys[layer_name]
        if input_pair not in input_keys_for_layer:
            raise ValueError(
                f"Could not unregister side input {input_pair} for layer "
                f"{layer_name}.  This input does not exist."
            )
        input_keys_for_layer.remove(input_pair)

    def checkpoint_activations(
        self, preserve_rng_state: bool = True, use_reentrant: bool = False
    ) -> None:
        """
        Enables activation checkpointing for the forward pass.  After this method is
        called, the full forward sequence executes using
        ``torch.utils.checkpoint.checkpoint()``.  Use :meth:`.store_activations` to
        undo this change.

        Args:
            preserve_rng_state (:obj:`bool`): The ``preserve_rng_state`` argument for
                ``torch.utils.checkpoint.checkpoint()``.
            use_reentrant (:obj:`bool`): The ``use_reentrant`` argument for
                ``torch.utils.checkpoint.checkpoint()``.
        """
        self._activation_checkpointing_options.is_enabled = True
        self._activation_checkpointing_options.preserve_rng_state = preserve_rng_state
        self._activation_checkpointing_options.use_reentrant = use_reentrant
        _logger.debug(
            "Enabled activation checkpointing for Sequential layer "
            "(preserve_rng_state=%s, use_reentrant=%s)",
            preserve_rng_state,
            use_reentrant,
        )

    def store_activations(self) -> None:
        """
        Reverses the effect of :meth:`checkpoint_activations`.
        """
        self._activation_checkpointing_options.is_enabled = False

    # pylint: disable-next=all
    def forward(self, input, **side_inputs):
        """
        Performs a forward pass.

        Each layer receives as input (1) the output from the previous layer, which is
        unpacked if the output is a :obj:`tuple` and ``unpack_tuple_inputs`` is
        ``True``, and (2) a mapped subset of keyword arguments from ``side_inputs``,
        where the subset for each layer is controlled by the layer's
        ``side_input_keys``.  Specifically, the layer receives a side input if the
        keyword passed to :meth:`Sequential.forward` is in the layer's list of
        ``side_input_keys``.

        If the layer has no side outputs, then the method returns the output of the
        final layer.  Otherwise, the method returns a :class:`OutputWithSideOutputs`.
        The ``output`` attribute contains the output of the final layer, and the
        ``side_outputs`` entry contains a :obj:`dict` of side outputs.  The keys of this
        :obj:`dict` correspond to output names (specified by ``side_output_keys``), and
        the values correspond to outputs of the specified layers.
        """

        if self.has_residual_connection:
            side_inputs["residual_input"] = input
        if not self._activation_checkpointing_options.is_enabled:
            return self._compute_forward_sequence(input, **side_inputs)

        options = self._activation_checkpointing_options

        return _torch_utils_checkpoint.checkpoint(
            self._compute_forward_sequence,
            input,
            preserve_rng_state=options.preserve_rng_state,
            use_reentrant=options.use_reentrant,
            **side_inputs,
        )

    # pylint: disable-next=all
    def _compute_forward_sequence(self, input, **side_inputs):
        side_outputs = {}
        output = input

        for layer_name, layer in self.named_layers:
            layer_kwargs = self._map_side_inputs_for_layer(side_inputs, layer_name)
            should_unpack_input = isinstance(output, tuple) and self.unpack_tuple_inputs

            if should_unpack_input:
                output = layer(*output, **layer_kwargs)
            else:
                output = layer(output, **layer_kwargs)

            if isinstance(output, _side_outputs.OutputWithSideOutputs):
                side_outputs = _side_outputs.merge_side_outputs(
                    side_outputs, output.side_outputs
                )
                output = output.output
            if layer_name in self.side_output_keys:
                key = self.side_output_keys[layer_name]
                side_outputs = _side_outputs.merge_side_outputs(
                    side_outputs, {key: output}
                )

        if len(side_outputs) > 0:
            return _side_outputs.OutputWithSideOutputs(output, side_outputs)
        return output

    def _map_side_inputs_for_layer(
        self, side_inputs: _Dict[str, _Any], layer_name: str
    ) -> _Dict[str, _Any]:
        mapped_side_inputs = {}
        keys_for_layer = self.side_input_keys.get(layer_name, [])
        for parent_input_name, child_input_name in keys_for_layer:
            if parent_input_name not in side_inputs:
                continue
            if child_input_name.startswith("**"):
                mapped_side_inputs.update(side_inputs[parent_input_name])
            else:
                mapped_side_inputs[child_input_name] = side_inputs[parent_input_name]
        return mapped_side_inputs

    def group_subsequences_by_lengths(self, lengths: _List[int]) -> None:
        """
        Reorganizes the ``nn.Module`` graph without affecting the output of the forward
        pass.  In particular, this method replaces contiguous subsequences of child
        layers with one :class:`Sequential` child per subsequence.  This provides
        flexibility to wrap arbitrary subsequences of layers with training optimizations
        such as FSDP.

        Args:
            lengths (:obj:`list` of :obj:`int`): A list that specifies the length of
                each contiguous subsequence to group.  This arg must contain strictly
                positive values, and it must sum to the total number of layers.
        """

        if any(length <= 0 for length in lengths):
            raise ValueError(
                "Input to group_subsequences_by_lengths contains a non-positive value"
            )
        boundaries = [0] + _helpers.cumsum(lengths)
        lengths_sum = boundaries[-1]
        if lengths_sum != len(self):
            raise ValueError(
                f"Input to group_subsequences_by_lengths sums to {lengths_sum}, which "
                f"is not the length of the Sequential instance ({len(self)})"
            )
        original_copy = self[:]
        del self[:]
        self._unpack_tuple_inputs = False
        for start_idx, end_idx in zip(boundaries[:-1], boundaries[1:]):
            layer = original_copy[start_idx:end_idx]
            group_idx = len(self)
            name = f"group_{group_idx}"
            side_input_keys = set()
            for input_keys_for_sublayer in layer.side_input_keys.values():
                new_keys = [key for key, _ in input_keys_for_sublayer]
                side_input_keys.update(new_keys)
            side_input_keys = [(key, key) for key in side_input_keys]
            self.append(layer, name=name, side_input_keys=side_input_keys)

    @_warnings.deprecate(alternative="group_subsequences_by_lengths")
    def group_children_by_sizes(self, sizes: _List[int]) -> None:
        """Deprecated.  Please use group_subsequences_by_lengths instead."""
        return self.group_subsequences_by_lengths(sizes)

    def rename_layers(self, new_names: _List[str]) -> None:
        """
        Renames all of the layers in the sequence.

        Args:
            new_names (:obj:`list` of :obj:`str`): The new names for each layer.  The
                names must be unique, and if the sequence has a residual connection, its
                name must remain ``'residual_connection'``.
        """
        if len(new_names) != len(self):
            raise ValueError(
                "The length of new_names differs from the number of layers"
            )
        if len(set(new_names)) != len(new_names):
            raise ValueError("new_names is not unique")

        old2new = dict(zip(self.layer_names, new_names))

        if self.has_residual_connection:
            if old2new["residual_connection"] != "residual_connection":
                raise ValueError(
                    "The name of 'residual_connection' must remain "
                    "'residual_connection'"
                )
        elif "residual_connection" in new_names:
            raise ValueError(
                "new_names cannot contain 'residual_connection' when there is no "
                "residual connection layer"
            )

        self_copy = self[:]
        input_keys_copy = self.side_input_keys.copy()
        output_keys_copy = self.side_output_keys.copy()

        del self[:]
        for idx, name in enumerate(new_names):
            self.add_module(name, self_copy[idx])

        self._side_input_keys = {  # pylint: disable=attribute-defined-outside-init
            old2new[key]: value for key, value in input_keys_copy.items()
        }

        self._side_output_keys = {
            old2new[key]: value for key, value in output_keys_copy.items()
        }

    def append(
        self,
        module: _nn.Module,
        *,
        name: _Optional[str] = None,
        side_input_keys: _Optional[_Union[_List[str], _List[_Tuple[str, str]]]] = None,
        side_output_key: _Optional[str] = None,
    ) -> "Sequential":
        """
        Appends ``module`` to the end of the sequence.

        Args:
            module (:obj:`nn.Module`): The module to append.
            name (:obj:`str`): The module's name.
            side_input_keys (:obj:`list` of :obj:`str`, optional): A list of side input
                names that ``module`` accepts.
            side_output_key (:obj:`str`, optional): A name for the output if the
                ``module`` produces a side output.
        """

        if name is None:
            name = f"layer_{len(self)}"
        name = _helpers.make_key_unique(name, self.layer_names)
        self.add_module(name, module)
        if side_input_keys is not None:
            side_input_keys = [
                (key, key) if isinstance(key, str) else key for key in side_input_keys
            ]
            self.side_input_keys[name] = side_input_keys
        if side_output_key is not None:
            self.side_output_keys[name] = side_output_key
        return self

    def insert(
        self,
        index: int,
        module: _nn.Module,
        *,
        name: _Optional[str] = None,
        side_input_keys: _Optional[_List[str]] = None,
        side_output_key: _Optional[str] = None,
    ):
        """
        Inserts ``module`` into the sequence before position ``index``.

        Args:
            index (:obj:`int`): The insert position.
            module (:obj:`nn.Module`): The module to append.
            name (:obj:`str`): The module's name.
            side_input_keys (:obj:`list` of :obj:`str`, optional): A list of side input
                names that ``module`` accepts.
            side_output_key (:obj:`str`, optional): A name for the output if the
                ``module`` produces a side output.
        """

        right_side = self[index:]
        del self[index:]
        self.append(
            module,
            name=name,
            side_input_keys=side_input_keys,
            side_output_key=side_output_key,
        )
        self += right_side
        return self

    def extend(self, sequential) -> "Sequential":
        """
        Extends the sequence by appending layers from another sequential layer.

        Args:
            sequential (:obj:`Sequential` or :obj:`torch.nn.Sequential`): The other
                sequential module that contains layers to append.
        """

        self += sequential
        return self

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            selected_modules = dict(list(self.named_layers)[idx])
            side_input_keys = {
                key: self.side_input_keys[key].copy()
                for key in selected_modules
                if key in self.side_input_keys
            }
            side_output_keys = {
                key: self.side_output_keys[key]
                for key in selected_modules
                if key in self.side_output_keys
            }
            return Sequential(
                selected_modules,
                side_input_keys=side_input_keys,
                side_output_keys=side_output_keys,
                unpack_tuple_inputs=self.unpack_tuple_inputs,
            )
        if isinstance(idx, str):
            return self._modules[idx]
        return _helpers.get_item_by_index(self.layers, idx)

    def __setitem__(self, idx, value):
        if isinstance(idx, slice):
            selected_keys = list(self.layer_names)[idx]
            if len(selected_keys) != len(value):
                raise ValueError(
                    f"slice specifies {len(selected_keys)} layers but value has length "
                    f"{len(value)}"
                )
            for key, module in zip(selected_keys, value):
                self.add_module(key, module)
        elif isinstance(idx, str):
            self.add_module(idx, value)
        else:
            key = _helpers.get_item_by_index(self.layer_names, idx)
            self.add_module(key, value)

    def __delitem__(self, idx):
        if isinstance(idx, slice):
            selected_keys = list(self.layer_names)[idx]
            for key in selected_keys:
                delattr(self, key)
                self.side_input_keys.pop(key, None)
                self.side_output_keys.pop(key, None)
        elif isinstance(idx, str):
            delattr(self, idx)
            self.side_input_keys.pop(idx, None)
            self.side_output_keys.pop(idx, None)
        else:
            key = _helpers.get_item_by_index(self.layer_names, idx)
            delattr(self, key)
            self.side_input_keys.pop(key, None)
            self.side_output_keys.pop(key, None)

    def __len__(self) -> int:
        return len(self._modules)

    def _add(self, other, in_place=False) -> "Sequential":
        result = self if in_place else self[:]
        if isinstance(other, Sequential):
            if self.unpack_tuple_inputs != other.unpack_tuple_inputs:
                raise ValueError(
                    "Add operator requires objects with the same value for "
                    "unpack_tuple_inputs but this is not the case "
                    f"({self.unpack_tuple_inputs} vs. {other.unpack_tuple_inputs})"
                )
            for name, layer in other.named_layers:
                side_input_keys = other.side_input_keys.get(name)
                side_input_keys = _helpers.copy_if_not_none(side_input_keys)
                side_output_key = other.side_output_keys.get(name)
                result.append(
                    layer,
                    name=name,
                    side_input_keys=side_input_keys,
                    side_output_key=side_output_key,
                )
        elif isinstance(other, _collections.abc.Mapping):
            for name, layer in other.items():
                result.append(layer, name=name)
        else:
            for layer in other:
                result.append(layer)
        return result

    def __add__(self, other) -> "Sequential":
        return self._add(other, in_place=False)

    def __radd__(self, other) -> "Sequential":
        result = self[:0]
        result += other
        result += self
        return result

    def __iadd__(self, other) -> "Sequential":
        return self._add(other, in_place=True)

    def _mul(self, other: int, in_place: bool = False) -> "Sequential":
        if not isinstance(other, int):
            raise TypeError(
                f"Unsupported operand types for *: {type(self)} and {type(other)}"
            )
        if other < 0:
            raise ValueError(
                f"Non-positive multiplication factor {other} for {type(self)}"
            )
        original_copy = self[:]
        result = self if in_place else self[:]
        del result[:]
        for _ in range(other):
            result += original_copy
        return result

    def __mul__(self, other: int) -> "Sequential":
        return self._mul(other, in_place=False)

    def __rmul__(self, other: int) -> "Sequential":
        return self * other

    def __imul__(self, other: int) -> "Sequential":
        return self._mul(other, in_place=True)

    def __iter__(self):
        return iter(self.layers)



"""
Provides a registry class for organizing a library of factory functions.
"""

import collections as _collections
import dataclasses as _dataclasses
import typing as _typing
import warnings as _warnings
from typing import Any as _Any
from typing import Callable as _Callable
from typing import Dict as _Dict
from typing import List as _List
from typing import Optional as _Optional
from typing import Tuple as _Tuple
from typing import Union as _Union

from tamm import _helpers

RegistrySpec = _collections.namedtuple(
    "RegistrySpec", ["key", "args", "kwargs"], defaults=["default", (), {}]
)


@_dataclasses.dataclass
class RegistryEntry:
    """Data for each entry in the registry"""

    factory_fn: _Callable
    description: str
    is_deprecated: bool = False


class Registry:
    """
    Organizes a library of factory functions, providing a one-stop spot to create
    instances of objects from the library and to discover members of the library.

    Args:
        name (:obj:`str`): A name for the registry.
    """

    def __init__(self, name: str = ""):
        self._name = name
        self._entries: _Dict[str, _Tuple[_Callable, str]] = {}

    @property
    def name(self) -> str:
        """The registry's name."""
        return self._name

    def register(
        self,
        factory_fn: _Optional[_Callable] = None,
        key: str = "default",
        description: _Optional[str] = None,
        is_deprecated: bool = False,
    ) -> None:
        """
        Adds an entry to the registry.

        Args:
            factory_fn (a :obj:`Callable` or `None`): If a :obj:`Callable`, the method
                saves this function to the registry.  If `None`, the method returns
                a decorator that saves the wrapped :obj:`Callable` to the registry.
            key (:obj:`str`): A key for accessing the `factory_fn` in the registry.
                Logs a warning if an entry already has this key.
            description (:obj:`str`): A text description of the entry.
            is_deprecated (:obj:`bool`): Flag for deprecating the entry.  Defaults to
                ``False``.

        Returns:
            `None` except when `factory_fn` is `None`, in which case the method returns
            a decorator.

        Examples:

            Basic usage:

                .. code-block:: python

                    class VisionTransformer:
                        def __init__(self, num_layers: int = 48):
                            ...

                    registry.register(
                        factory_fn=VisionTransformer,
                        key="vit",
                        description="VisionTransformer model."
                    )

            Decorator usage:

                .. code-block:: python

                    @registry.register(key="vit", description="Vision transformer model.")
                    def vit_factory(num_layers: int = 48) -> VisionTransformer:
                        return VisionTransformer(num_layers=num_layers)

        """
        if factory_fn is None:
            return self._get_decorator_for_register(
                key=key, description=description, is_deprecated=is_deprecated
            )

        self._validate_name_string(calling_function_name=str(factory_fn), name=key)
        if key in self._entries:
            _warnings.warn(
                f"Overwriting existing entry for key {key} in registry {self.name}"
            )
        if not callable(factory_fn):
            raise ValueError(
                f"Registry {self.name} received a non-callable factory function for "
                f"key '{key}'"
            )

        if description is None:
            description = _helpers.get_cleaned_docstring(factory_fn)
        else:
            description = _helpers.clean_docstring(description)

        self._entries[key] = RegistryEntry(
            factory_fn=factory_fn, description=description, is_deprecated=is_deprecated
        )
        return None

    def _get_decorator_for_register(
        self, *, key: str, description: str, is_deprecated: bool
    ) -> _Callable:
        """
        Returns a decorator that takes a factory_fn as input, adds the factory_fn to
        the registry, and the returns the unmodified factory_fn.
        """

        def decorator(factory_fn):
            self.register(
                factory_fn,
                key=key,
                description=description,
                is_deprecated=is_deprecated,
            )
            return factory_fn

        return decorator

    @staticmethod
    def _validate_name_string(
        calling_function_name: str, name: _Optional[str] = None
    ) -> None:
        """
        Validates that the provided name is of type string.

        Args:
            calling_function_name (str): Name of the function from where this
            method is being called. This is used to generate the error message.
            name (Optional[str]): Name to validate.

        Raises ValueError: If `name` is not of type str.
        """
        if not isinstance(name, str):
            raise ValueError(
                f"{calling_function_name} expects a str for the name argument "
                f"but received a {type(name)} instead."
            )

    @_typing.overload
    def get_factory_fn(self, key: _Union[list, tuple]) -> _Callable:
        ...

    def get_factory_fn(self, key: str) -> _Callable:
        """
        Retrieves a factory function from the registry.

        Args:
            key (:obj:`str` or obj:`tuple`): The key specified when adding the entry to
            the registry.  If a :obj:`tuple`, the first element must be the ``key`` as
            a :obj:`str` and optional second and/or third elements may be an ``arg``
            tuple and/or ``kwarg`` dictionary for the factory function.

        Returns:
            The factory function associated with `key`.
        """
        key, *_ = self._unpack_key(key)
        if key not in self._entries:
            raise KeyError(
                f"Registry {self.name} has no entry with key {key} "
                f"(existing keys: {sorted([*self._entries])})"
            )
        return self._entries[key].factory_fn

    def _unpack_key(self, key: _Union[str, list, tuple]) -> tuple:
        if not isinstance(key, (tuple, list)):
            return key, (), {}
        if len(key) not in (1, 2, 3):
            raise ValueError(
                f"Registry {self.name} received a tuple input of length {len(key)}, "
                "but tuple keys must have length 1, 2, or 3."
            )
        new_key = key[0]
        args = key[1] if len(key) > 1 else ()
        if isinstance(args, _collections.abc.Mapping):
            kwargs, args = args, ()
        else:
            kwargs = key[2] if len(key) > 2 else {}
        return new_key, args, kwargs

    @_typing.overload
    def create(self, key: _Union[list, tuple]) -> _Any:
        ...

    def create(self, key: str, *args: _Any, **kwargs: _Any) -> _Any:
        """
        Returns a new object created with the factory function associated with `key`.

        Args:
            key (:obj:`str` or :obj:`tuple`): The key (a :obj:`str`) corresponding to
                the registered factory function for creating the return object.
                Alternatively, this can also be a :obj:`tuple`---in this case, the first
                element must be the ``key`` as a :obj:`str` and optional second and/or
                third elements may be a packed ``arg`` tuple and/or ``kwarg`` dictionary
                for the factory function.
            *args, **kwargs: Positional and keyword arguments for the factory function
                associated with `key`.

        Returns:
            The result of `factory_fn(*args, **kwargs)`, where `factory_fn` is the
            factory function associated with `key`.
        """
        key, args_from_key, kwargs_from_key = self._unpack_key(key)
        if not args:
            args = args_from_key
        kwargs = {**kwargs_from_key, **kwargs}
        factory_fn = self.get_factory_fn(key)
        return factory_fn(*args, **kwargs)  # pylint: disable=not-callable

    def is_valid_key(self, obj: _Any) -> bool:
        """
        Returns ``True`` if ``obj`` corresponds to an entry in the registry when passed
        as a ``key`` argument to :meth:`.create`.  Returns ``False`` otherwise.
        """
        try:
            key, *_ = self._unpack_key(obj)
            return key in self._entries
        except Exception:
            return False

    def get_entries(self) -> _Dict[str, RegistryEntry]:
        """
        Returns entries of the registry as a dictionary that maps each key to a
        corresponding :obj:`RegistryEntry`.
        """
        return self._entries.copy()

    def list(
        self,
        include_descriptions: bool = False,
        filter_deprecated: bool = False,
    ) -> _Union[_List[str], _Dict[str, str]]:
        """
        List the keys in the registry.

        Args:
            include_descriptions (bool): Whether to include descriptions for each key.

        Returns:
            List of keys if `include_descriptions` is False.
            Dictionary of keys and their descriptions if `include_descriptions` is True.
        """
        entries = self.get_entries()
        if filter_deprecated:
            entries = {k: v for k, v in entries.items() if not v.is_deprecated}
        keys_and_entries = sorted(entries.items())
        if not include_descriptions:
            return [key for key, _ in keys_and_entries]
        return {key: entry.description for key, entry in keys_and_entries}

    def print(
        self,
        *,
        include_descriptions: bool = False,
        filter_deprecated: bool = False,
        wide: bool = False,
    ):
        """
        Print the keys and descriptions in the registry.

        Args:
            include_descriptions: Whether to include descriptions for each key.
            filter_deprecated: Whether to remove deprecated entries.
            wide: Flag to disable output truncation.
        """
        entries = self.list(
            include_descriptions=include_descriptions,
            filter_deprecated=filter_deprecated,
        )
        if include_descriptions:
            max_len_names = max(len(name) for name in entries)
            paddings = (" " * (max_len_names + 4 - len(name)) for name in entries)
            lines = [
                name + padding + descr
                for padding, (name, descr) in zip(paddings, entries.items())
            ]
        else:
            lines = entries
        if not wide:
            lines = _helpers.truncate_lines_for_terminal(*lines)
        print("\n".join(lines))



import copy as _copy
from typing import Dict as _Dict
from typing import Optional as _Optional

from tamm.layers.transformer.layer_sequence import vanilla as _vanilla
from tamm.typing import ModuleOrBuilder as _ModuleOrBuilder


class UniformTransformerLayerSequence(_vanilla.TransformerLayerSequence):
    """
    A generic sequence of transformer layers.  Unlike :class:`TransformerLayerSequence`,
    the architecture remains consistent across layers.

    Args:
        layer_template: A template layer or builder.  The sequence layer deep copies
            this template to initialize itself.
        num_layers (:obj:`int`): The number of layers in the sequence.
        output_hidden_states (:obj:`bool`): A flag for including outputs from each
            layer in the sequence output.  Defaults to ``False``.
        attention_types (:obj:`dict`, optional): A dictionary that
            maps layer indices to their secondary attention types.  For example
            if ``attention_types={3: "global", 7: "global"},
            then layers 3 and 7 receive the arguments
            ``attention_side_inputs=secondary_attention_side_inputs["global"]``
            rather than ``attention_side_inputs=attention_side_inputs``.
    """

    def __init__(
        self,
        layer_template: _ModuleOrBuilder,
        *,
        num_layers: int = 1,
        output_hidden_states: bool = False,
        attention_types: _Optional[_Dict[int, str]] = None,
    ):
        sequence = [_copy.deepcopy(layer_template) for _ in range(num_layers)]
        for layer in sequence:
            try:
                layer.reset_parameters()
            except AttributeError:
                pass
        super().__init__(
            *sequence,
            output_hidden_states=output_hidden_states,
            attention_types=attention_types,
        )



import random as _random
from dataclasses import dataclass as _dataclass
from typing import List as _List
from typing import Optional as _Optional
from typing import Tuple as _Tuple
from typing import Union as _Union

import torch as _torch
from torch import nn as _nn

from tamm.generation.token_generators import TokenGenerator as _TokenGenerator
from tamm.generation.token_generators.base import (
    TokenGeneratorOutput as _TokenGeneratorOutput,
)
from tamm.generation.token_generators.next_token_logic import (
    NextTokenLogic as _NextTokenLogic,
)
from tamm.generation.token_generators.utils import (
    KVCacheType,
    create_random_number_generator,
    find_eos_index,
    model_forward_step,
    pad_after_termination,
    post_process_token_ids,
    setup_kv_cache,
)
from tamm.layers.transformer.kv_cache import BaseKVCache as _BaseKVCache


@_dataclass
class SpeculativeDecodingTokenGeneratorOutput(_TokenGeneratorOutput):
    """
    Output generated from :py:class:`SpeculativeDecodingTokenGenerator`, consisting of token_ids,
    number of suggested tokens by the draft generator and number of accepted draft tokens for each
    sequence in the input batch.

    Attributes:
        token_ids (:py:class:`_torch.Tensor`): Tokens produced by the speculative decoding generator.
          Does not include ``input_ids``.
        suggested_token_counts (:py:class:`_torch.Tensor`): Total number of tokens suggested by the draft model.
          The tensor has shape ``(batch_size, N)``, where ``N`` is the number of times draft generator
          is called for speculative decoding.
        accepted_token_counts (:py:class:`_torch.Tensor`): Total number of tokens which were generated by
         the draft model and accepted as final tokens. The tensor has shape ``(batch_size, N)``,
         where ``N`` is the number of times draft generator is called for speculative decoding.
    """

    token_ids: _torch.Tensor
    suggested_token_counts: _torch.Tensor
    accepted_token_counts: _torch.Tensor


# pylint: disable=too-many-locals
class SpeculativeDecodingTokenGenerator(_TokenGenerator):
    """Token generator that uses speculative decoding strategy. Based on the paper:
    `Fast Inference from Transformers via Speculative Decoding <https://arxiv.org/pdf/2211.17192>`_

    Args:
        model (:obj:`nn.Module`): The target model instance from which tokens need to be sampled.
        next_token_logic (:obj:`_NextTokenLogic`): Logic specifying how target model's output logits should be
         used to generate the next token.
        draft_generator (:obj:`_TokenGenerator`): Token generator for producing draft tokens to be used for
          speculative decoding. draft_generator should provide an option called output_probabilities. When this
          option is set to ``True``, the draft generator should return probabilities used for generating the tokens,
          along with the generated tokens.
        max_draft_tokens (:obj:`int`): The maximum number of draft tokens generated in each round of speculative
          decoding. This argument is called ``gamma`` in the paper.
        cache_type (:obj:`KVCacheType`, optional): Type of KV cache to use. Defaults to ``KVCacheType.Speculative``.
          Only supported types are ``KVCacheType.Speculative`` and ``None``.
        cache_dtype (:obj:`torch.dtype`, optional): Data type for KV cache. Defaults to ``None``.
    """

    def __init__(
        self,
        model: _nn.Module,
        next_token_logic: _NextTokenLogic,
        draft_generator: _TokenGenerator,
        max_draft_tokens: int,
        cache_type: _Optional[KVCacheType] = KVCacheType.SPECULATIVE,
        cache_dtype: _Optional[_torch.dtype] = None,
    ):
        super().__init__(
            model=model,
            eos_id=draft_generator.eos_ids,
            pad_id=draft_generator.pad_id,
        )
        self._verify_draft_generator_compatibility(draft_generator)
        self._draft_generator = draft_generator
        self._next_token_logic = next_token_logic
        self._cache_type = cache_type
        self._cache_dtype = cache_dtype
        self._first_sd_call = True
        self.max_draft_tokens = max_draft_tokens

    def _verify_draft_generator_compatibility(self, draft_generator: _TokenGenerator):
        assert hasattr(draft_generator, "output_probabilities"), (
            "draft_generator should provide an option called output_probabilities to return probabilities used for "
            f"generation, to be compatible with {self.__class__.__name__}."
        )

    @_torch.inference_mode()
    def generate(
        self,
        input_ids: _torch.Tensor,
        max_new_tokens: _Union[int, _List[int]],
        seed: _Optional[int] = None,
    ) -> SpeculativeDecodingTokenGeneratorOutput:
        token_ids = input_ids.clone()

        bs, prompt_length = token_ids.shape
        device = token_ids.device

        prng = _random.Random(seed)

        if isinstance(max_new_tokens, list):
            tokens_until_max = _torch.tensor(max_new_tokens, device=device)
        else:
            assert isinstance(max_new_tokens, int)
            tokens_until_max = _torch.full((bs,), max_new_tokens, device=device)

        max_max_new_tokens = max_new_tokens  # max across all batch sequences
        if isinstance(max_new_tokens, list):
            max_max_new_tokens = max(max_new_tokens)

        decode_terminator = [x.to(device) for x in self.eos_ids]

        all_suggested = _torch.zeros((bs, 0), device=device)
        all_accepted = _torch.zeros((bs, 0), device=device)

        should_terminate = _torch.zeros((bs,), device=device, dtype=_torch.bool)

        target_cache = setup_kv_cache(
            model=self.model,
            batch_size=bs,
            cache_len=prompt_length + max_max_new_tokens,
            input_ids=input_ids,
            pad_id=self.pad_id,
            cache_type=self._cache_type,
            cache_dtype=self._cache_dtype,
        )

        self._first_sd_call = True

        while not should_terminate.all().item():
            # we set num_speculative_tokens based on number of expected tokens to reach max_new_tokens
            # we use max_remaining_tokens - 1 because last token comes from target model

            max_remaining_tokens = tokens_until_max.amax().item()
            num_speculative_tokens = min(
                self.max_draft_tokens, max_remaining_tokens - 1
            )

            # using no persistent draft cache, so we always pass in full input for next round
            output = self._speculative_decode_once(
                token_ids,
                num_speculative_tokens,
                target_cache,
                random_number_generator=prng,
            )

            token_ids = _torch.concat([token_ids, output.token_ids], dim=-1)

            non_pad_tokens = (output.token_ids != self.pad_id).sum(dim=-1)
            tokens_until_max -= non_pad_tokens

            # ignore counts if a sequence has terminated
            output.suggested_token_counts[should_terminate] = 0
            output.accepted_token_counts[should_terminate] = 0

            all_suggested = _torch.concat(
                [all_suggested, output.suggested_token_counts[:, None]],
                dim=-1,
            )
            all_accepted = _torch.concat(
                [all_accepted, output.accepted_token_counts[:, None]],
                dim=-1,
            )

            should_terminate_round = self._should_terminate_decoding(
                tokens_until_max,
                decode_terminator,
                output.token_ids,
            )

            should_terminate |= should_terminate_round

            if should_terminate.all().item():
                break

        token_ids = post_process_token_ids(
            token_ids[:, prompt_length:], max_new_tokens, self.pad_id
        )
        return SpeculativeDecodingTokenGeneratorOutput(
            token_ids=pad_after_termination(token_ids, self.pad_id, decode_terminator),
            suggested_token_counts=all_suggested.to(_torch.int32),
            accepted_token_counts=all_accepted.to(_torch.int32),
        )

    def _speculative_decode_once(
        self,
        token_ids: _torch.Tensor,
        num_speculative_tokens: int,
        target_cache: _Optional[_BaseKVCache],
        random_number_generator: _random.Random,
    ) -> SpeculativeDecodingTokenGeneratorOutput:
        """
        Generates num_speculative_tokens speculative tokens from draft generator and 1 token from
        the target model. Uses rejection sampling to accept n <= num_speculative_tokens speculative tokens.

        Args:
            token_ids (:obj:`_torch.Tensor`): Input token ids for the draft generator. Since we do not use
              a persistent cache for graft generation, we pass all the tokens generated so far to the draft
              generator.
            num_speculative_tokens (:obj:`int`): Number of tokens to generate from draft generator.
            target_cache: (:obj:`_BaseKVCache` or obj:`None`): KV cache for the target model.
            random_number_generator (:obj:`_random.Random`): A pseudo random number generator used to
              generating seeds used for sampling form draft generator and target model's probability distribution.
        """
        device = token_ids.device
        batch_size = token_ids.shape[0]

        target_input_ids = token_ids
        if target_cache and not self._first_sd_call:
            # New sample generated from target in last round needs to be fed to target
            target_input_ids = target_input_ids[:, -1:]

        if num_speculative_tokens == 0:
            target_logits = self._compute_target_logits(
                target_input_ids,
                target_prompt_len=token_ids.shape[-1],
                draft_token_count=0,
                target_cache=target_cache,
                with_temporary_cache=False,
            )
            final_token_ids = self._next_token_logic(
                target_logits[:, -1],
                rng=create_random_number_generator(
                    device=device, seed=random_number_generator.getrandbits(32)
                ),
                return_probabilities=False,
            ).token_ids

            return SpeculativeDecodingTokenGeneratorOutput(
                token_ids=final_token_ids,
                suggested_token_counts=_torch.zeros((batch_size,), device=device),
                accepted_token_counts=_torch.zeros((batch_size,), device=device),
            )

        # using no persistent cache, so always pass in full input to draft model;
        # at output, we get newly generated tokens and their logits
        self._draft_generator.output_probabilities = True  # type: ignore[attr-defined]

        draft_output = self._draft_generator.generate(
            token_ids,
            num_speculative_tokens,
            seed=random_number_generator.getrandbits(32),
        )
        draft_tokens = draft_output.token_ids
        draft_probs = draft_output.probabilities  # type: ignore[attr-defined]

        draft_token_count = draft_tokens.shape[1]
        # draft_token_count can be less than num_speculative_tokens if draft has early terminated in
        # all elements in the batch
        early_termination_amount = (draft_tokens == self.pad_id).sum(dim=-1)

        # append draft token ids to target token ids
        target_input_ids = _torch.cat([target_input_ids, draft_tokens], dim=-1)

        # on first target call, we treat the last target_input_id as if it was generated during the "last" target call.
        # because from second target call onwards, target_input_id always have an extra token generated
        # during the last call prepended, which is not present in the target cache but is fed to the target model
        target_prompt_len = token_ids.shape[-1] - 1

        target_logits = self._compute_target_logits(
            target_input_ids,
            target_prompt_len=target_prompt_len,
            draft_token_count=draft_token_count,
            target_cache=target_cache,
            with_temporary_cache=True,
        )

        (
            target_token,
            num_accepted_tokens,
        ) = self._verify_from_target(
            draft_tokens,
            draft_probs,
            target_logits,
            draft_token_count,
            random_number_generator,
        )  # evolve initial seed deterministically on every round

        max_accepted_tokens = _torch.max(num_accepted_tokens).item()

        # merge temporary cache into main cache, +1 because we want to merge cache values corresponding
        # to the token generated by target model from previous round
        if target_cache:
            target_cache.merge(
                merge_length=num_accepted_tokens + 1,
                max_merge_length=max_accepted_tokens + 1,
                previous_length=target_prompt_len,
            )

        accepted = draft_tokens.clone()
        # align sequences across batch dimension
        accepted[
            _torch.arange(draft_token_count, device=device)
            >= num_accepted_tokens[:, None]
        ] = self.pad_id
        accepted = accepted[:, :max_accepted_tokens]
        final_token_ids = _torch.concat([accepted, target_token], dim=-1)

        suggested_count = draft_token_count - early_termination_amount
        accepted_count = _torch.min(num_accepted_tokens, suggested_count)

        self._first_sd_call = False

        return SpeculativeDecodingTokenGeneratorOutput(
            token_ids=final_token_ids,
            suggested_token_counts=suggested_count,
            accepted_token_counts=accepted_count,
        )

    def _compute_target_logits(
        self,
        target_input_ids: _torch.Tensor,
        target_prompt_len: int,
        draft_token_count: int,
        target_cache: _Optional[_BaseKVCache],
        with_temporary_cache: bool,
    ) -> _torch.Tensor:
        if self._first_sd_call and target_cache:
            # pre-fill cache
            model_forward_step(
                model=self.model,
                input_ids=target_input_ids[:, :target_prompt_len],
                next_tokens=None,
                pad_id=self.pad_id,
                kv_cache=target_cache,
                prompt_len=target_prompt_len,
                token_idx=None,
            )
            # remove all pre-filled tokens
            target_input_ids = target_input_ids[:, target_prompt_len:]
        if target_cache and with_temporary_cache:
            target_cache.create_temporary_cache_view(
                speculation_length=draft_token_count + 1
            )
        output = model_forward_step(
            model=self.model,
            input_ids=target_input_ids,
            next_tokens=None,
            pad_id=self.pad_id,
            kv_cache=target_cache,
            prompt_len=target_prompt_len,
            token_idx=None,
        )
        # draft_token_count tokens from draft and one token from target
        # output shape: batch_size x draft_token_count + 1 x vocab_size
        return output.logits[
            :, -(draft_token_count + 1) :, : self.model.config.vocab_size
        ]

    def _verify_from_target(
        self,
        draft_tokens: _torch.Tensor,
        draft_probs: _torch.Tensor,
        target_logits: _torch.Tensor,
        draft_token_count: int,
        random_number_generator: _random.Random,
    ) -> _Tuple[_torch.Tensor, _torch.Tensor]:
        device = draft_tokens.device

        target_probs = self._next_token_logic.compute_probabilities(
            target_logits
        )  # shape: batch_size x draft_token_count + 1 x vocab_size

        # sample next token assuming all draft tokens are accepted
        # we do this to make selection of acceptance tokens without if/else loops
        # this token is always rejected, since we resample from new_probs later
        target_sample_seed = random_number_generator.getrandbits(32)
        target_token = self._sample_from_target_probs(
            target_probs[:, -1], seed=target_sample_seed
        )  # shape: batch_size x 1

        all_draft_tokens = _torch.concat(
            [draft_tokens[:, -draft_token_count:].clone(), target_token],
            dim=1,
        )  # shape: batch_size x draft_token_count + 1

        # add inf to the end of draft probs, this is to align draft prob shape
        # with the target_probs shape
        draft_probs = _torch.concat(
            [
                draft_probs,
                _torch.inf * _torch.ones_like(draft_probs[..., -1, :]).unsqueeze(1),
            ],
            dim=-2,
        )

        # get the acceptance probability for every logit
        # since draft_probs is inf at the last index, acceptance_prob is 0
        acceptance_prob = _torch.clamp(target_probs / draft_probs, max=1.0)

        # find the acceptance probability for the suggested logits
        acceptance_prob = _torch.gather(
            acceptance_prob, -1, all_draft_tokens[..., None]
        ).squeeze(
            -1
        )  # shape: batch_size x draft_token_count + 1

        # roll the die
        roll_seed = random_number_generator.getrandbits(32)
        roll = _torch.rand(
            acceptance_prob.shape,
            generator=create_random_number_generator(device=device, seed=roll_seed),
            device=acceptance_prob.device,
        )

        # since acceptance_prob is 0 at the last index, at least
        # one token (the one from target) would be rejected
        rejected = roll > acceptance_prob

        num_accepted_tokens = rejected.to(_torch.int32).argmax(dim=-1)

        # resample from adjusted target distribution
        # new_probs will be zero at the last index (corresponding to target token)
        new_probs = _torch.clamp(target_probs - draft_probs, min=0)
        new_probs /= _torch.sum(new_probs, keepdim=True, dim=-1)
        # when all draft tokens are accepted, since we set new_probs to be same as
        # target_probs at the last index, we end up drawing from the target_probs, which
        # is what we want
        new_probs[:, -1, :] = target_probs[:, -1, :]

        new_probs = new_probs[
            _torch.arange(new_probs.shape[0], device=device), num_accepted_tokens, :
        ]

        target_token = self._sample_from_target_probs(
            new_probs, seed=target_sample_seed
        )

        return target_token, num_accepted_tokens

    @staticmethod
    def _should_terminate_decoding(
        tokens_until_max: _torch.Tensor,
        decode_terminator: _List[_torch.Tensor],
        token_ids: _torch.Tensor,
    ) -> _torch.Tensor:
        max_gen_termination = tokens_until_max <= 0

        termination_indices = find_eos_index(token_ids, decode_terminator)
        eos_reached = termination_indices < _torch.iinfo(_torch.int32).max

        should_terminate = _torch.logical_or(eos_reached, max_gen_termination)

        return should_terminate

    @staticmethod
    def _sample_from_target_probs(probs: _torch.Tensor, seed: int) -> _torch.Tensor:
        rng = create_random_number_generator(device=probs.device, seed=seed)
        return _torch.multinomial(probs, num_samples=1, generator=rng)



"""
layers.basic
^^^^^^^^^^^^

This submodule contains simple miscellaneous layers.

.. autoclass:: tamm.layers.ChannelsFirstToLast
    :show-inheritance:

.. autoclass:: tamm.layers.ChannelsLastToFirst
    :show-inheritance:

.. autoclass:: tamm.layers.Index
    :show-inheritance:

.. autoclass:: tamm.layers.Interpolation
    :show-inheritance:
    :members: forward

.. autoclass:: tamm.layers.Mean
    :show-inheritance:

.. autoclass:: tamm.layers.SelectByKey
    :show-inheritance:
    :members: forward

.. autoclass:: tamm.layers.SoftCap
    :members: forward

.. autoclass:: tamm.layers.Union
    :show-inheritance:
"""

from typing import Any as _Any
from typing import Dict as _Dict
from typing import Tuple as _Tuple
from typing import Union as _Union

import torch as _torch
import torch.nn as _nn

from tamm import _helpers
from tamm.layers import functional as _tamm_F
from tamm.layers.common import LayerMixin as _LayerMixin
from tamm.typing import OptionalModuleOrBuilder as _OptionalModuleOrBuilder


class ChannelsFirstToLast(_nn.Module, _LayerMixin):
    """
    A layer that converts channels-first tensors with shape
    ``(batch_size, num_channels, *spatial_dims)`` to tensors with shape
    ``(batch_size, *spatial_dims, num_channels)``.

    Args:
        flatten_spatial_dims (:obj:`bool`): A flag for flattening the spatial dimensions
            into a single dimension when ``True``.  Defaults to ``False``.
    """

    def __init__(self, flatten_spatial_dims: bool = False):
        super().__init__()
        self.flatten_spatial_dims = flatten_spatial_dims

    def extra_repr(self):
        return f"flatten_spatial_dims={self.flatten_spatial_dims}"

    def forward(self, tensor: _torch.Tensor) -> _torch.Tensor:
        if self.flatten_spatial_dims:
            tensor = tensor.flatten(start_dim=2)
        return tensor.movedim(1, -1)


class ChannelsLastToFirst(_nn.Module, _LayerMixin):
    """
    A layer that converts channels-last tensors with shape
    ``(batch_size, *spatial_dims, num_channels)`` to tensors with shape
    ``(batch_size, num_channels, *spatial_dims)``.
    """

    def __init__(self):
        super().__init__()

    def forward(self, tensor: _torch.Tensor) -> _torch.Tensor:
        return tensor.movedim(-1, 1)


class Index(_nn.Module, _LayerMixin):
    """
    A layer for indexing tensors with predetermined indices, slices, etc.

    Args:
        expressions: A sequence of expressions for indexing inputs to the layer with the
            ``[]`` operator.  This can include slices, integers, lists of integers,
            ``None``, ``...``, etc.

    Example:

        .. code-block:: python

            layer = Index(slice(1, None), ..., -1, None)
            out = layer(tensor)   # equivalent to out = tensor[1:, ..., -1, None]
    """

    def __init__(self, *expressions):
        super().__init__()
        self.expressions = expressions

    def extra_repr(self):
        pieces = [_helpers.maybe_slice_to_string(expr) for expr in self.expressions]
        return ", ".join(pieces)

    def forward(self, x: _torch.Tensor) -> _torch.Tensor:
        return x[self.expressions]


class Interpolation(_nn.Module, _LayerMixin):
    """
    A layer for resizing spatial tensors using :func:`torch.nn.functional.interpolate`.
    The layer expects inputs in channels-first format.

    Args:
        mode (:obj:`str`): The interpolation mode, such as ``"nearest"`` or
            ``"bicubic"``.  In addition to the options provided by
            :func:`torch.nn.functional.interpolate`, this layer also supports
            ``"pillow-style-bilinear"`` and ``"pillow-style-bicubic"``, which match the
            behavior of Pillow (the ``recompute_scale_factor`` and ``antialias`` options
            are ignored in these cases).
        cast_dtype (:obj:`torch.dtype`, optional): The dtype to perform the
            interpolation in.  Set to ``None`` to use the input dtype.  Defaults to
            ``float32``.

    Please see :func:`torch.nn.functional.interpolate` for descriptions of the remaining
    options.
    """

    PILLOW_BICUBIC = "pillow-style-bicubic"
    PILLOW_BILINEAR = "pillow-style-bilinear"

    def __init__(
        self,
        mode: str = "nearest",
        size: _Union[_Tuple[int, ...], int, None] = None,
        scale_factor: _Union[_Tuple[float, ...], float, None] = None,
        align_corners: _Union[bool, None] = None,
        recompute_scale_factor: _Union[bool, None] = None,
        antialias: bool = False,
        cast_dtype: _Union[_torch.dtype] = _torch.float32,
    ):
        super().__init__()
        self.mode = mode
        self.size = size
        self.scale_factor = scale_factor
        self.align_corners = align_corners
        self.recompute_scale_factor = recompute_scale_factor
        self.antialias = antialias
        self.cast_dtype = cast_dtype

    @property
    def _options(self):
        return {
            "mode": self.mode,
            "size": self.size,
            "scale_factor": self.scale_factor,
            "align_corners": self.align_corners,
            "recompute_scale_factor": self.recompute_scale_factor,
            "antialias": self.antialias,
        }

    @property
    def _interpolate_kwargs(self):
        kwargs = self._options
        mode = kwargs["mode"]
        if mode in (self.PILLOW_BICUBIC, self.PILLOW_BILINEAR):
            kwargs["antialias"] = True
            kwargs["align_corners"] = False
            if mode == self.PILLOW_BILINEAR:
                kwargs["mode"] = "bilinear"
            elif mode == self.PILLOW_BICUBIC:
                kwargs["mode"] = "bicubic"
            else:
                raise RuntimeError(f"Pillow-style mode '{mode}' not recognized")
        return kwargs

    def extra_repr(self):
        options = self._options
        if not options["antialias"]:
            options.pop("antialias")
        if self.mode in (self.PILLOW_BICUBIC, self.PILLOW_BILINEAR):
            options.pop("antialias", None)  # these are ignored in this case
            options.pop("align_corners")
        strings = [
            f"{key}={repr(val)}" for key, val in options.items() if val is not None
        ]
        return ", ".join(strings)

    def forward(
        self,
        inputs: _torch.Tensor,
        size: _Union[_Tuple[int, ...], int, None] = None,
        scale_factor: _Union[_Tuple[float, ...], float, None] = None,
    ) -> _torch.Tensor:
        """
        Args:
            inputs (:obj:`torch.Tensor`): The input tensor to interpolate.  This must
                have shape ``(batch_size, dim, *spatial_shape)``.
            size (:obj:`tuple` of :obj:`int`, optional): The spatial shape of the
                output tensor.  Defaults to ``None``.
            scale_factor (:obj:`tuple` of :obj:`float`, optional): The spatial shape of
                the output tensor relative to that of ``inputs``.  Defaults to ``None``.

        .. note::
            If both ``size`` and ``scale_factor`` are ``None``, then these arguments
            default to the layer's ``size`` and ``scale_factor`` attributes.

        Returns:
            An interpolation of ``inputs``.
        """

        input_dtype = inputs.dtype
        if self.cast_dtype is not None:
            inputs = inputs.type(self.cast_dtype)

        kwargs = self._interpolate_kwargs
        if size is not None or scale_factor is not None:
            kwargs["size"] = size
            kwargs["scale_factor"] = scale_factor

        with _helpers.autocast_disabled(inputs.device):
            return _torch.nn.functional.interpolate(inputs, **kwargs).type(input_dtype)


class Map(_nn.Module, _LayerMixin):
    """
    A layer that applies a child layer separately to every tensor in the layer's inputs.

    Args:
        func:  A :obj:`.LayerBuilder` or :obj:`nn.Module` that specifies the mapping
            function.

    Example:

        .. code-block:: python

            func = tamm.layers.Lambda(lambda x: x + 10)
            layer = Map(func)

            x = {"a": torch.tensor(1.0), "b": torch.tensor(-1.0)}
            y = layer(x)  # {'a': tensor(11.), 'b': tensor(9.)}
            z = layer(list(y.values()))  # [tensor(21.), tensor(19.)]
    """

    def __init__(self, func: _OptionalModuleOrBuilder):
        super().__init__()
        _helpers.append_children(self, func=func)

    def forward(self, inputs: _Any) -> _Dict[str, _Any]:
        from torch.utils import _pytree  # pylint: disable=import-outside-toplevel

        return _pytree.tree_map_only(_torch.Tensor, self.func, inputs)


class Mean(_nn.Module, _LayerMixin):
    """
    A layer that applies :func:`torch.mean` to its inputs.

    Args:
        dim (:obj:`int` or :obj:`tuple`, optional): The ``dim`` arg for
            :func:`torch.mean`.  By default, the mean is across the last dimension
            (``-1``).
        keepdim (:obj:`bool`, optional): The ``keepdim`` arg for
            :func:`torch.mean`.  Defaults to ``False``.
    """

    def __init__(self, dim=-1, keepdim=False):
        super().__init__()
        self.dim = dim
        self.keepdim = keepdim

    def forward(self, x):
        return _torch.mean(x, dim=self.dim, keepdim=self.keepdim)

    def extra_repr(self):
        return f"dim={self.dim}, keepdim={self.keepdim}"


class Sum(_nn.Module, _LayerMixin):
    """
    A layer that applies :func:`torch.sum` to its inputs.

    Args:
        dim (int): Dimension to sum across.
    """

    def __init__(self, dim=-1, keepdim=False):
        super().__init__()
        self.dim = dim
        self.keepdim = keepdim

    def forward(self, x):
        return _torch.sum(x, dim=self.dim, keepdim=self.keepdim)

    def extra_repr(self):
        return f"dim={self.dim}, keepdim={self.keepdim}"


class SelectByKey(_nn.Module, _LayerMixin):
    """
    A layer that returns a specific value from an input :obj:`dict`.

    Args:
        key: A dictionary key.
    """

    def __init__(self, key: _Any):
        super().__init__()
        self.key = key

    def extra_repr(self):
        return repr(self.key)

    def forward(self, inputs: _Dict[_Any, _Any]) -> _Any:
        """
        Args:
            inputs (:obj:`dict`): A dictionary that contains ``key`` as a key.

        Returns:
            The value ``inputs[key]``.

        Raises:
            KeyError: If ``inputs`` does not contain the key.
        """
        return inputs[self.key]


class SoftCap(_nn.Module, _LayerMixin):
    """
    Computes ``cap * tanh(x / cap)``, which is a smooth and differentiable way
    to cap the values of a tensor.

    Args:
        cap (:obj:`float`): The value to cap inputs to.
    """

    def __init__(self, cap: float):
        super().__init__()
        self.cap = cap

    def extra_repr(self):
        return f"cap={self.cap}"

    def forward(self, tensor: _torch.Tensor) -> _torch.Tensor:
        return _tamm_F.soft_cap(tensor, cap=self.cap)


class Union(_nn.Module, _LayerMixin):
    """
    A layer for composing child layers that take the same inputs and produce different
    outputs.  The layer returns a :obj:`dict` that maps the name of each child layer
    to its output.

    Args:
        named_layers (:obj:`dict`): A dictionary that maps the name of each segmentation
            layer to a :obj:`.LayerBuilder` or :obj:`nn.Module`.
    """

    def __init__(self, named_layers: _Dict[str, _OptionalModuleOrBuilder]):
        super().__init__()
        _helpers.append_children(self, **named_layers)

    def forward(self, inputs: _Any) -> _Dict[str, _Any]:
        return {name: child(inputs) for name, child in self.named_children()}


class PadToMultiple(_nn.Module, _LayerMixin):
    """
    A layer that pads the specified dimension(s) of a tensor with a given value so that its size becomes
    a multiple of a specified number. The padding is applied to the end (right side) of the specified dimensions.

    Args:
        multiple (:obj:`int`): The number to which the specified dimension(s) should be a multiple of.
        dim (:obj: `Union[int, Tuple[int, ...]]`): The dimension(s) to pad. Defaults to ``-1``.
        pad_value (:obj: `float`): The value to use for padding. Defaults to ``0``.


    Example:
        >>> layer = PadToMultiple(multiple=3, dim=(-2, -1), pad_value=0)
        >>> x = torch.randn(1, 3, 14, 14)
        >>> y = layer(x)
        >>> print(y.shape)  # Output shape will be (1, 3, 15, 15) since 15 is the next multiple of 3
    """

    def __init__(
        self,
        multiple: int,
        *,
        dim: _Union[int, _Tuple[int, ...]] = -1,
        pad_value: float = 0,
    ):
        super().__init__()
        self.multiple = multiple
        self.dim = dim if isinstance(dim, tuple) else (dim,)
        self.pad_value = pad_value

    def extra_repr(self):
        return f"multiple={self.multiple}, dim={self.dim}, pad_value={self.pad_value}"

    def forward(self, x: _torch.Tensor) -> _torch.Tensor:
        pad = []

        normalized_dims = set(dim % x.ndim for dim in self.dim)
        min_dim = min(normalized_dims)

        any_pad = False
        # Ensure we pad in reversed order when multiple dims are given since nn.functional.pad applies in reverse order
        for d in reversed(range(min_dim, x.ndim)):
            if d in normalized_dims:
                pad_num = (-x.size(d)) % self.multiple
            else:
                pad_num = 0

            if pad_num > 0:
                pad.extend([0, pad_num])  # Right padding
                any_pad = True
            else:
                pad.extend([0, 0])

        if any_pad:
            x = _nn.functional.pad(x, pad, value=self.pad_value)

        return x



"""
layers.mixture_of_experts
-------------------------

This module implements building blocks for mixture of experts layers.


Main layer
^^^^^^^^^^

.. autoclass:: tamm.layers.mixture_of_experts.MixtureOfExperts


Router
^^^^^^

.. autoclass:: tamm.layers.mixture_of_experts.MixtureOfExpertsRouter

.. autoclass:: tamm.layers.mixture_of_experts.router.MixtureOfExpertsRouterOutput


Samplers
^^^^^^^^

.. autoclass:: tamm.layers.mixture_of_experts.sampler.MixtureOfExpertsSampler
    :members: forward

.. autoclass:: tamm.layers.mixture_of_experts.sampler.MixtureOfExpertsSamplerOutput

.. autoclass:: tamm.layers.mixture_of_experts.TopKMixtureOfExpertsSampler
    :show-inheritance:


Dispatcher
^^^^^^^^^^

.. autoclass:: tamm.layers.mixture_of_experts.dispatcher.MixtureOfExpertsDispatcher
    :members: forward

.. autoclass:: tamm.layers.mixture_of_experts.DroplessMixtureOfExpertsDispatcher
    :show-inheritance:

.. autoclass:: tamm.layers.mixture_of_experts.CapacityConstrainedMixtureOfExpertsDispatcher
    :show-inheritance:

.. autoclass:: tamm.layers.mixture_of_experts.dispatcher.MixtureOfExpertsDispatcherForwardMode
    :members:

.. autoclass:: tamm.layers.mixture_of_experts.dispatcher.MixtureOfExpertsDispatchOutput
    :members:


Capacity constraint
^^^^^^^^^^^^^^^^^^^


.. autoclass:: tamm.layers.mixture_of_experts.dispatcher.MixtureOfExpertsCapacityConstraint
    :members: forward

.. autoclass:: tamm.layers.mixture_of_experts.VanillaMixtureOfExpertsCapacityConstraint
    :show-inheritance:

.. autoclass:: tamm.layers.mixture_of_experts.dispatcher.MixtureOfExpertsCapacityConstraintOutput


Losses
^^^^^^

.. autoclass:: tamm.layers.mixture_of_experts.LoadBalanceLoss
    :members: forward

.. autoclass:: tamm.layers.mixture_of_experts.RouterZLoss
    :members: forward
"""

from tamm.layers.mixture_of_experts.dispatcher import (
    CapacityConstrainedMixtureOfExpertsDispatcher,
    DroplessMixtureOfExpertsDispatcher,
    VanillaMixtureOfExpertsCapacityConstraint,
)
from tamm.layers.mixture_of_experts.loss import LoadBalanceLoss, RouterZLoss
from tamm.layers.mixture_of_experts.mixture_of_experts import MixtureOfExperts
from tamm.layers.mixture_of_experts.router import MixtureOfExpertsRouter
from tamm.layers.mixture_of_experts.sampler import TopKMixtureOfExpertsSampler

__all__ = [
    "CapacityConstrainedMixtureOfExpertsDispatcher",
    "DroplessMixtureOfExpertsDispatcher",
    "LoadBalanceLoss",
    "MixtureOfExpertsRouter",
    "MixtureOfExperts",
    "RouterZLoss",
    "TopKMixtureOfExpertsSampler",
    "VanillaMixtureOfExpertsCapacityConstraint",
]



# pylint: disable=abstract-method

import abc as _abc
import warnings as _warnings
from typing import Optional as _Optional
from typing import Tuple as _Tuple

import torch as _torch
import transformers as _transformers
from transformers import modeling_outputs as _modeling_outputs

import tamm as _tamm
from tamm.hf.transformers import modeling_utils as _modeling_utils
from tamm.hf.transformers.models.causal_lm import kv_cache as _kv_cache
from tamm.hf.transformers.models.causal_lm.configuration_causal_lm import (
    TammCausalLMConfig as _TammCausalLMConfig,
)
from tamm.typing import StateDictType as _StateDictType


class TammCausalLMPreTrainedModel(_modeling_utils.PreTrainedModel):
    """
    A base class for wrapping |tamm| :class:`CausalLMTransformer` models
    into a HuggingFace model type.
    """

    config_class = _TammCausalLMConfig
    base_model_prefix = "transformer"
    supports_gradient_checkpointing = True
    _no_split_modules = ["TransformerDecoderLayer"]

    @classmethod
    def from_tamm_config(
        cls,
        config: _tamm.models.common.ModelConfig,
        state_dict: _Optional[_StateDictType] = None,
        **kwargs,
    ) -> "TammCausalLMPreTrainedModel":
        """
        Creates and returns a Hugging Face-compatible wrapper model.

        Args:
            config: A |tamm| model config for a :obj:`.CausalLMTransformer`.
            state_dict: An optional state dict for the model.  If ``None``,
                the model states are loaded from the config's pretrained path
                (if provided) and otherwise initialized from scratch.
            **kwargs: Optional keyword arguments to pass to
                :class:`TammCausalLMConfig`.
        """

        hf_config = _TammCausalLMConfig(tamm_config=config, **kwargs)
        with _torch.device("meta"):
            hf_model = cls(hf_config)

        if state_dict is None:
            tamm_model = config.create_model(pretrained=True)
            state_dict = tamm_model.state_dict()
        hf_model.get_tamm_model().load_state_dict(state_dict, assign=True)

        hf_model.tie_weights()

        return hf_model

    @classmethod
    def from_tamm_model(cls, model: _torch.nn.Module) -> "TammCausalLMPreTrainedModel":
        """
        Creates and returns a Hugging Face-compatible wrapper model.

        Args:
            model: The |tamm| model to wrap into a Hugging Face type.
        """

        hf_config = _TammCausalLMConfig()
        with _torch.device("meta"):
            hf_model = cls(hf_config)

        hf_model.set_tamm_model(model)

        hf_model.tie_weights()

        return hf_model

    @_abc.abstractmethod
    def get_tamm_model(self) -> _torch.nn.Module:
        """Get the underlying tamm model"""

    @_abc.abstractmethod
    def set_tamm_model(self, model: _torch.nn.Module) -> None:
        """Set the underlying tamm model"""

    def _init_weights(self, module):
        # Hugging Face calls this for every submodule
        if hasattr(module, "reset_parameters"):
            module.reset_parameters()

    def tie_weights(self):
        # pylint: disable=assignment-from-none
        model = self.get_tamm_model()
        should_tie = getattr(model, "is_embedding_tied_to_output_transform", False)
        if not should_tie:
            return

        output_embeddings = self.get_output_embeddings()
        if output_embeddings is None:
            return
        self._tie_or_clone_weights(output_embeddings, self.get_input_embeddings())

        for module in self.modules():
            if hasattr(module, "_tie_weights"):
                module._tie_weights()  # pylint: disable=protected-access

    def gradient_checkpointing_enable(self, gradient_checkpointing_kwargs=None):
        if gradient_checkpointing_kwargs is None:
            gradient_checkpointing_kwargs = {}
        tamm_model = self.get_tamm_model()
        for layer in tamm_model.layers:
            layer.checkpoint_activations(**gradient_checkpointing_kwargs)

    def gradient_checkpointing_disable(self):
        tamm_model = self.get_tamm_model()
        for layer in tamm_model.layers:
            layer.store_activations()

    def _maybe_warn_about_ignored_args_to_forward(self, **ignored_args):
        output_attentions = ignored_args.pop("output_attentions", False)
        if output_attentions and not self.config.output_attentions:
            _warnings.warn(
                "CausalLM forward() will ignore output_attentions=True. Please "
                "pass output_attentions=True instead to the tamm model config."
            )
        output_hidden_states = ignored_args.pop("output_hidden_states", False)
        if output_hidden_states and not self.config.output_hidden_states:
            _warnings.warn(
                "CausalLM forward() will ignore output_hidden_states=True. Please "
                "pass output_hidden_states=True instead to the tamm model config."
            )
        for key, value in ignored_args.items():
            if value is None:
                continue
            _warnings.warn(
                f"The CausalLM forward() method received a value for {key}, but "
                "this arg will be ignored.  Please contact the tamm developers "
                "to prioritize support for this feature."
            )

    def _should_use_return_dict(self, return_dict):
        return self.config.use_return_dict if return_dict is None else return_dict


class TammCausalLMModel(TammCausalLMPreTrainedModel):
    def __init__(self, config: _TammCausalLMConfig):
        super().__init__(config)
        self.config = config
        self.tamm_model = None

        if config.tamm_config is not None:
            tamm_model = config.tamm_config.create_model()
        else:
            _warnings.warn("TammCausalLMModel created without a tamm model")
            tamm_model = None
        self.set_tamm_model(tamm_model)

        self.post_init()

    def get_tamm_model(self) -> _torch.nn.Module:
        return self.tamm_model

    def set_tamm_model(self, model: _torch.nn.Module) -> None:
        self.tamm_model = model

    def get_input_embeddings(self):
        return self.tamm_model.embedding

    def set_input_embeddings(self, value):
        self.tamm_model.embedding = value

    def forward(
        self,
        input_ids=None,
        *,
        attention_mask=None,
        past_key_values: _Optional[_Tuple[_Tuple[_torch.Tensor]]] = None,
        inputs_embeds: _Optional[_torch.FloatTensor] = None,
        use_cache: _Optional[bool] = None,
        return_dict=None,
        **ignored_args,
    ):
        self._maybe_warn_about_ignored_args_to_forward(**ignored_args)
        inputs, mode = self._get_inputs_and_mode_for_forward(
            input_ids=input_ids, inputs_embeds=inputs_embeds
        )
        if attention_mask is None:
            attention_mask = _torch.ones(
                *inputs.shape[:2], dtype=_torch.int32, device=inputs.device
            )
        kv_cache_view = self._get_kv_cache_view(
            past_key_values=past_key_values,
            attention_mask=attention_mask,
            use_cache=use_cache,
        )
        if use_cache:
            attention_mask = attention_mask[:, -inputs.size(1) :]
        output = self.tamm_model(
            inputs,
            segment_ids=attention_mask,
            kv_cache=kv_cache_view,
            mode=mode,
        )
        return self._create_forward_return_value(
            tamm_output=output,
            kv_cache_view=kv_cache_view,
            return_dict=return_dict,
        )

    @staticmethod
    def _get_inputs_and_mode_for_forward(*, input_ids, inputs_embeds):
        if input_ids is not None:
            if inputs_embeds is not None:
                raise ValueError("Forward received both input_ids and inputs_embeds")
            return (
                input_ids,
                "skip_output_layer",
            )  # Hugging Face uses a separate LM head layer
        if inputs_embeds is not None:
            return (inputs_embeds, "skip_embedding_and_output_layers")
        raise ValueError("Forward received neither input_ids nor inputs_embeds")

    def _get_kv_cache_view(self, *, past_key_values, attention_mask, use_cache):
        if not use_cache:
            return None
        positions = attention_mask.cumsum(dim=1).sub_(1).clamp_min_(0)
        if past_key_values is None:
            num_layers = self.get_tamm_model().num_layers
            return _kv_cache.HFStyleKVCacheView.create_empty_cache(
                num_layers=num_layers, segment_ids=attention_mask, positions=positions
            )
        return _kv_cache.HFStyleKVCacheView(
            past_key_values, segment_ids=attention_mask, positions=positions
        )

    def _create_forward_return_value(
        self,
        *,
        tamm_output,
        kv_cache_view: _kv_cache.HFStyleKVCacheView,
        return_dict: bool,
    ):
        if kv_cache_view is not None:
            past_key_values = kv_cache_view.past_key_values
        else:
            past_key_values = None

        output = {
            "last_hidden_state": tamm_output.last_hidden_state,
            "hidden_states": tamm_output.hidden_states,
            "attentions": tamm_output.attentions,
            "past_key_values": past_key_values,
        }

        if not self._should_use_return_dict(return_dict):
            return tuple(val for val in output.values() if val is not None)
        return _modeling_outputs.BaseModelOutputWithPast(**output)

    def _reorder_cache(
        self, past_key_values: _Tuple[_Tuple[_torch.Tensor]], beam_idx: _torch.Tensor
    ):
        beam_idx = beam_idx.to(past_key_values[0][0].device)
        return tuple(
            tuple(tensor.index_select(0, beam_idx) for tensor in layer)
            for layer in past_key_values
        )


class TammCausalLMForCausalLM(
    TammCausalLMPreTrainedModel, _transformers.GenerationMixin
):
    def __init__(self, config: _TammCausalLMConfig):
        super().__init__(config)
        self.transformer = TammCausalLMModel(config)

        self.lm_head = None
        if self.get_tamm_model() is not None:
            self._init_lm_head()

        self.post_init()

    def _init_lm_head(self):
        model = self.get_tamm_model()
        if model is None:
            self.lm_head = None
            return

        output_transform = model.output_transform
        hidden_dim = output_transform.in_features
        vocab_size = output_transform.out_features
        device = output_transform.weight.device
        dtype = output_transform.weight.dtype

        self.lm_head = _torch.nn.Linear(
            hidden_dim, vocab_size, bias=False, device=device, dtype=dtype
        )
        self.lm_head.load_state_dict(output_transform.state_dict())

    @property
    def _tied_weights_keys(self):
        keys = [*super()._tied_weights_keys]

        model = self.get_tamm_model()
        try:
            has_tied_embedding = self.lm_head.weight is model.embedding.weight
        except AttributeError:
            has_tied_embedding = False

        if has_tied_embedding:
            keys.append("lm_head.weight")

        return keys

    def get_tamm_model(self) -> _torch.nn.Module:
        return self.transformer.get_tamm_model()

    def set_tamm_model(self, model: _torch.nn.Module) -> None:
        self.transformer.set_tamm_model(model)
        self._init_lm_head()
        self._maybe_update_generation_config()

    def _maybe_update_generation_config(self):
        if self.generation_config is None:
            return
        tamm_model = self.get_tamm_model()
        if tamm_model is None:
            return
        try:
            tokenizer = _tamm.create_tokenizer(tamm_model.metadata.tokenizer_spec)
        except Exception:
            _warnings.warn(
                "Could not create a tokenizer to update the HuggingFace generation config. If "
                "generating text, please ensure that the eos_token_id and pad_token_id are set correctly."
            )
            return

        if hasattr(tokenizer, "bos_id"):
            self.generation_config.bos_token_id = tokenizer.bos_id

        if hasattr(tokenizer, "eot_id"):
            self.generation_config.eos_token_id = tokenizer.eot_id
        elif hasattr(tokenizer, "eos_id"):
            self.generation_config.eos_token_id = tokenizer.eos_id
        else:
            _warnings.warn(
                "Could not update the eos_token_id in the HuggingFace generation config. "
                "If generating text, please ensure that the eos_id is set correctly."
            )

        if hasattr(tokenizer, "pad_id"):
            self.generation_config.pad_token_id = tokenizer.pad_id
        else:
            _warnings.warn(
                "Could not update the pad_token_id in the HuggingFace generation config. "
                "If generating text, please ensure that the pad id is set correctly."
            )

    def get_output_embeddings(self):
        return self.lm_head

    def set_output_embeddings(self, new_embeddings):
        self.lm_head = new_embeddings

    def prepare_inputs_for_generation(
        self,
        input_ids: _torch.LongTensor,
        past_key_values: _Optional[_Tuple[_Tuple[_torch.Tensor]]] = None,
        attention_mask: _Optional[_torch.Tensor] = None,
        inputs_embeds: _Optional[_torch.FloatTensor] = None,
        cache_position: _Optional[_torch.LongTensor] = None,
        **kwargs,
    ):
        try:
            return super().prepare_inputs_for_generation(
                input_ids=input_ids,
                past_key_values=past_key_values,
                attention_mask=attention_mask,
                inputs_embeds=inputs_embeds,
                cache_position=cache_position,
                **kwargs,
            )
        except NotImplementedError:
            # transformers.GenerationMixin only implements this method since
            # v4.46--when not implemented, fall back to old implementation
            pass

        if past_key_values is not None:
            input_ids = input_ids[:, -1:]

        if inputs_embeds is not None:
            if past_key_values is None:
                input_ids = None
            else:
                inputs_embeds = None

        return {
            "input_ids": input_ids,
            "attention_mask": attention_mask,
            "past_key_values": past_key_values,
            "inputs_embeds": inputs_embeds,
            "cache_position": cache_position,
            **kwargs,
        }

    def forward(
        self,
        input_ids: _Optional[_torch.Tensor] = None,
        *,
        attention_mask: _Optional[_torch.Tensor] = None,
        inputs_embeds: _Optional[_torch.FloatTensor] = None,
        past_key_values: _Optional[_Tuple[_Tuple[_torch.Tensor]]] = None,
        use_cache: _Optional[bool] = None,
        labels=None,
        return_dict=None,
        **ignored_args,
    ):
        self._maybe_warn_about_ignored_args_to_forward(**ignored_args)
        backbone_output = self.transformer(
            input_ids,
            attention_mask=attention_mask,
            inputs_embeds=inputs_embeds,
            past_key_values=past_key_values,
            use_cache=use_cache,
            return_dict=True,
        )
        logits = self.lm_head(backbone_output.last_hidden_state)
        loss = self._compute_loss(logits, labels)
        return self._create_forward_return_value(
            backbone_output=backbone_output,
            logits=logits,
            loss=loss,
            return_dict=return_dict,
        )

    @staticmethod
    def _compute_loss(logits, labels):
        if labels is None:
            return None
        logits = logits[..., :-1, :]
        labels = labels[..., 1:]
        logits = logits.reshape(-1, logits.shape[-1])
        labels = labels.flatten()
        return _torch.nn.functional.cross_entropy(logits, labels)

    def _create_forward_return_value(
        self, *, backbone_output, logits, loss, return_dict
    ):
        output = {
            "loss": loss,
            "logits": logits,
            "hidden_states": backbone_output.hidden_states,
            "attentions": backbone_output.attentions,
            "past_key_values": backbone_output.past_key_values,
        }
        if not self._should_use_return_dict(return_dict):
            return tuple(val for val in output.values() if val is not None)
        return _modeling_outputs.CausalLMOutputWithPast(**output)

    def _reorder_cache(
        self, past_key_values: _Tuple[_Tuple[_torch.Tensor]], beam_idx: _torch.Tensor
    ):
        # pylint: disable-next=protected-access
        return self.transformer._reorder_cache(past_key_values, beam_idx)



"""
This module implements the base converter type as well as some reusable components for
building converters of model state dicts.

"""

import abc as _abc
import logging as _logging
from typing import Optional as _Optional

import torch as _torch

from tamm import _helpers
from tamm.converters import common as _parent_common
from tamm.converters.models import registry as _registry
from tamm.typing import StateDictType as _StateDictType

_logger = _logging.getLogger(__name__)


class ModelStateDictConverter(_parent_common.StateDictConverter, _abc.ABC):
    """
    Interface for converting model state dicts.  Compared to the base
    :class:`StateDictConverter`, this class mainly adds methods for saving and loading
    model state.  This allows us to export to HuggingFace format, for example (for which
    must also save configs).  The class also adds :meth:`.from_tamm_state_dict` and
    :meth:`.from_other_state_dict`, which enable creation of converter instances by
    inferring the converter arguments from tensors and metadata in the state dict.
    """

    def __init_subclass__(cls, converter_id: _Optional[str] = None):
        """
        Recommended format for the converter_id is "{model-name}:{format}", using
        hyphen-separated (kebab) case for the model name and format.  For example,
        "deepseek:huggingface" converts a DeepSeek model to or from HuggingFace format.
        """
        if converter_id is None:
            return
        _registry.register_converter_cls(
            cls, converter_id=converter_id, description=cls.__doc__
        )

    def __init__(self, wrapped_converter):
        self._wrapped_converter = wrapped_converter

    def is_tamm_key(self, key: str) -> bool:
        return self._wrapped_converter.is_tamm_key(key)

    def is_other_key(self, key: str) -> bool:
        return self._wrapped_converter.is_other_key(key)

    def _convert_from_tamm_impl(self, state_dict: _StateDictType) -> _StateDictType:
        # pylint: disable-next=protected-access
        return self._wrapped_converter._convert_from_tamm_impl(state_dict)

    def _convert_to_tamm_impl(self, state_dict: _StateDictType) -> _StateDictType:
        # pylint: disable-next=protected-access
        return self._wrapped_converter._convert_to_tamm_impl(state_dict)

    @classmethod
    @_abc.abstractmethod
    def from_tamm_state_dict(cls, state_dict):
        """Returns an instance of the class, given a tamm state dict."""

    @classmethod
    @_abc.abstractmethod
    def from_other_state_dict(cls, state_dict):
        """Returns an instance of the class, given an "other" state dict."""

    @staticmethod
    def load_tamm_state_dict(filepath, **kwargs):
        """Load a tamm state dict from a file."""
        return _helpers.safe_torch_load(filepath, **kwargs)

    @staticmethod
    def save_tamm_state_dict(state_dict, filepath, **kwargs):
        """Save a tamm state dict to a file."""
        _torch.save(state_dict, filepath, **kwargs)

    @staticmethod
    @_abc.abstractmethod
    def load_other_state_dict(filepath):
        """Load a non-tamm state dict from a file."""

    @staticmethod
    @_abc.abstractmethod
    def save_other_state_dict(state_dict, filepath):
        """Save a non-tamm state dict to a file."""


class TorchModelConverter(ModelStateDictConverter):
    """Converter for non-tamm Torch models that use PyTorch state dicts."""

    @staticmethod
    def load_other_state_dict(filepath):
        return _helpers.safe_torch_load(filepath)

    @staticmethod
    def save_other_state_dict(state_dict, filepath):
        return _torch.save(state_dict, filepath)



from typing import TYPE_CHECKING
from typing import Dict as _Dict
from typing import Optional as _Optional

from tamm.layers.common.post_hooks.common import CompositePostHook, IdentityPostHook

if TYPE_CHECKING:
    from tamm.ao import ArchOptimizer


def ArchOptimizersPostHook(  # pylint: disable=invalid-name
    arch_optimizers: _Optional[_Dict[str, "ArchOptimizer"]] = None,
):
    if arch_optimizers is None:
        return IdentityPostHook()

    hooks = [ao.optimize for ao in arch_optimizers.values()]
    return CompositePostHook(*hooks)



"""
models.vision_transformer
=========================

.. autoclass:: tamm.models.VisionTransformer
    :show-inheritance:
    :members:

.. autoclass:: tamm.models.configs.VisionTransformerConfig
"""

from typing import Optional as _Optional
from typing import Tuple as _Tuple
from typing import Union as _Union

from tamm import layers as _layers
from tamm.models.common import ModelMixin as _ModelMixin


class VisionTransformer(_layers.VisionTransformerEncoder, _ModelMixin):
    """A transformer encoder for image inputs."""

    @classmethod
    def create_basic_builder(
        cls,
        *,
        hidden_dim: int = 768,
        num_layers: int = 12,
        num_heads: int = 12,
        hidden_dim_scale_factor: float = 4.0,
        input_dim: int = 3,
        patch_size: _Union[int, _Tuple[int, int]] = (16, 16),
        stride: _Union[int, _Tuple[int, int], None] = None,
        apply_positional_embedding: bool = True,
        positional_embedding_shape: _Optional[_Tuple[int, int]] = (14, 14),
        positional_embedding_interpolation: _Optional[str] = None,
        num_class_tokens: int = 0,
        activation: str = "relu",
        norm_eps: float = 1e-5,
        norm_bias: bool = False,
        attention_logits_soft_cap: _Optional[float] = None,
        output_transform: _Optional[_layers.common.LayerBuilder] = None,
        attention_dropout_p: float = 0.0,
        hidden_dropout_p: float = 0.0,
    ):
        """
        Args:
            hidden_dim (:obj:`int`): The hidden dim for the model's transformer layers.
            num_layers (:obj:`int`): The number of transformer layers.
            num_heads (:obj:`int`): The number of attention heads.
            hidden_dim_scale_factor (:obj:`float`, optional): For feed-forward layers,
                the ratio of the layer's hidden dimension to the layer's input
                dimension.
            input_dim (:obj:`int`): The input dimension (number of channels) for the
                model's image inputs.
            patch_size (:obj:`int` or pair of :obj:`int`): The kernel size for the
                :obj:`.ConvEmbedding` layer.
            stride (:obj:`int` or pair of :obj:`int`): The stride for the convolutional
                embedding.  If ``None``, this defaults to ``patch_size``.
            apply_positional_embedding (:obj:`bool`): A flag for including a
                :obj:`.SpatialPositionalEmbedding` in the embedding layer.  Defaults to
                ``True``.
            positional_embedding_shape (pair of :obj:`int`): The height and width of
                the positional embedding layer.
            positional_embedding_interpolation (:obj:`str`): An optional
                :class:`.Interpolation` mode (such as ``"bicubic"``) for rescaling the
                positional embeddings to match the shape of the input embeddings.
                Defaults to ``None``, in which case the input embedding shape must match
                the positional embeddings shape.
            num_class_tokens (:obj:`int`): The number of class tokens.
            activation (:obj:`str`, optional): The activation type.  Defaults to ``"relu"``.
            norm_eps (:obj:`float`): The ``eps`` value for norm layers.
            norm_bias (:obj:`bool`, optional): Whether to include a bias term in the
                normalization layer.  Defaults to ``False``.
            attention_logits_soft_cap (:obj:`float`, optional): An optional positive
                threshold for soft-capping attention logits.  Defaults to ``None``,
                which implies no capping.
            output_transform (:obj:`.LayerBuilder`, optional): An optional builder for
                the output transform layer.  Defaults to ``None``.
            attention_dropout_p (:obj:`float`, optional): The dropout rate for
                attention probabilities.  Defaults to 0.
            hidden_dropout_p (:obj:`float`, optional): The dropout rate for hidden
                states.
        """

        # pylint: disable=too-many-locals

        norm = _layers.LayerNorm.Builder(
            [hidden_dim],
            eps=norm_eps,
            bias=norm_bias,
        )

        class_token_embedding = _layers.ConstantEmbedding.Builder(
            num_embeddings=num_class_tokens,
            dim=hidden_dim,
        )
        conv_embedding = _layers.ConvEmbedding.create_builder(
            input_dim=input_dim,
            output_dim=hidden_dim,
            kernel_size=patch_size,
            stride=stride,
            apply_positional_embedding=apply_positional_embedding,
            positional_embedding_shape=positional_embedding_shape,
        )
        if apply_positional_embedding:
            conv_embedding.positional_encoding.interpolation = (
                positional_embedding_interpolation
            )
        conv_embedding.norm = norm
        embedding = _layers.UnionEmbedding.Builder(
            {
                "class_tokens": class_token_embedding,
                "convolution_tokens": conv_embedding,
            }
        )

        class_token_seg = _layers.ConstantSegmentation.Builder(num_class_tokens)
        conv_seg = (
            _layers.ConvEmbeddingPaddingTransform.create_builder_from_convolution_layer(
                conv_embedding.convolution
            )
        )
        segmentation_layers = {
            "class_tokens": class_token_seg,
            "convolution_tokens": conv_seg,
        }
        segmentation = _layers.UnionSegmentation(segmentation_layers)

        attention_mask = _layers.AttentionMask.Builder()

        attention = _layers.TransformerAttention.create_builder(
            target_dim=hidden_dim,
            norm=norm,
            num_heads=num_heads,
            attention_dropout_p=attention_dropout_p,
            hidden_dropout_p=hidden_dropout_p,
            qkv_transform_bias=True,
            output_transform_bias=True,
        )
        if attention_logits_soft_cap is not None:
            attention.scaled_dot_product_attention.logits_soft_cap = (
                attention_logits_soft_cap
            )

        feed_forward = _layers.TransformerFeedForward.create_builder(
            input_dim=hidden_dim,
            hidden_dim=round(hidden_dim * hidden_dim_scale_factor),
            norm=norm,
            activation=activation,
            activation_dropout_p=hidden_dropout_p,
            output_dropout_p=hidden_dropout_p,
            hidden_transform_bias=True,
            output_transform_bias=True,
        )

        layer = _layers.TransformerLayer.Builder(
            attention=attention, feed_forward=feed_forward
        )
        layers = _layers.UniformTransformerLayerSequence.Builder(
            layer, num_layers=num_layers
        )

        if isinstance(output_transform, _layers.common.LayerConfig):
            output_transform = output_transform.create_builder()

        return cls.Builder(
            embedding=embedding,
            segmentation=segmentation,
            attention_mask=attention_mask,
            layers=layers,
            output_norm=norm,
            output_transform=output_transform,
        )



from transformers import modeling_utils as _transformers_modeling_utils

from tamm import layers as _layers


class PreTrainedModel(_transformers_modeling_utils.PreTrainedModel):
    """
    Subclass of :class:`PreTrainedModel` from :mod:`transformers` for implementing bug
    fixes or additional functionality related to |tamm|.
    """

    # pylint: disable=abstract-method

    @property
    def _tied_weights_keys(self):
        tied_module_names = [
            name
            for name, module in self.named_modules()
            if isinstance(module, _layers.TiedWeightLinear)
        ]
        return [f"{name}.weight" for name in tied_module_names]



from typing import List, Union, cast

import tamm.utils.json as _tamm_json
from tamm.tokenizers.common import PublishedTokenizerConfig, Tokenizer, TokenizerConfig
from tamm.tokenizers.registry import TokenizerRegistry
from tamm.utils.uri.uri_handler import _URIHandler, is_uri_or_posix


class URIHandlerTokenizerRegistry(TokenizerRegistry):
    def _create_config(
        self, identifier: Union[str, "Tokenizer"], *args, **kwargs
    ) -> "TokenizerConfig":
        if not is_uri_or_posix(identifier):
            raise KeyError(
                f"{identifier} is neither a URL nor posix path, {self} cannot handle this"
            )

        with _URIHandler().open(identifier) as f:
            published_tokenizer_config = cast(
                PublishedTokenizerConfig, _tamm_json.load(f)
            )
        try:
            return published_tokenizer_config.tokenizer_config
        except AttributeError:
            return cast(TokenizerConfig, published_tokenizer_config)

    def list_(self, *args, **kwargs) -> List[str]:
        return []

    def describe(self, identifier: str, *args, **kwargs) -> str:
        if not is_uri_or_posix(identifier):
            raise KeyError(
                f"{identifier} is neither a URL nor posix path, {self} cannot handle this"
            )

        with _URIHandler().open(identifier) as f:
            published_tokenizer_config = _tamm_json.load(f)

        try:
            return published_tokenizer_config.description
        except AttributeError:
            return "<no_description>"



import contextlib as _contextlib
import hashlib as _hashlib
import shutil as _shutil
import tempfile as _tempfile
from pathlib import Path as _Path

from tamm.utils import user_dir_utils


def _get_dest_prefix_from_uri(uri: str) -> _Path:
    uri = str(uri)  # coerce uri into a string
    sha256_uri = _hashlib.sha256(uri.strip().encode("utf-8")).hexdigest()
    return _Path(user_dir_utils.get_tamm_cache_dir()) / sha256_uri[:2] / sha256_uri[2:]


def _get_dest_path_from_uri(uri: str) -> _Path:
    dest_prefix = _get_dest_prefix_from_uri(uri)
    basename = _Path(uri).name
    return dest_prefix / basename


@_contextlib.contextmanager
def atomic_download_prefix(dest_path: _Path):
    """
    Provides a temporary prefix for downloader to write into and then move the file
    over to the destination.

    Args:
        dest_path: Expected destination path
    """
    basename = dest_path.name
    dest_prefix_temp = _Path(_tempfile.mkdtemp(dir=user_dir_utils.get_tamm_tmp_dir()))
    dest_prefix_temp.mkdir(parents=True, exist_ok=True)
    yield dest_prefix_temp
    # Download to temp and then move to cache to ensure atomicity. PR #664
    dest_path.parent.mkdir(parents=True, exist_ok=True)
    _shutil.move(str(dest_prefix_temp / basename), dest_path)



"""
Compatibility helpers to support older versions of Pytorch
"""
import logging as _logging
from typing import Optional, Union

import torch as _torch

_logger = _logging.getLogger(__name__)


def _get_default_device() -> _torch.device:
    """
    This API resolves both
    1) PyTorch default device (`torch.get_default_device()`) and
    2) device context (i.e., `with torch.device(...)`)
    """

    return _torch.tensor([]).device


def _is_same_device_type(
    device1: Optional[Union[str, "_torch.device"]],
    device2: Optional[Union[str, "_torch.device"]],
) -> bool:
    """
    equality checks with a torch.device are a little confusing because they also have an
    index which can sometimes falsify the equality check unexpectedly.
    Whenever we want to condition on the device type, use the .type attribute
    (such as device.type == "meta")
    """
    if isinstance(device1, str):
        device1 = _torch.device(device1)
    if isinstance(device2, str):
        device2 = _torch.device(device2)
    return device1.type == device2.type


def _pytree_average(*trees, weights=None):
    from torch.utils import _pytree  # pylint: disable=import-outside-toplevel

    if len(trees) == 0:
        raise ValueError("pytree_average requires at least one tree")

    if weights is not None:
        if len(weights) != len(trees):
            raise ValueError(
                "pytree_average received a different number of weights than trees"
            )
        scale = len(weights) / sum(weights)
        weights = [x * scale for x in weights]

    flattened_trees = [_pytree.tree_flatten(tree) for tree in trees]
    leaves_list, specs_list = zip(*flattened_trees)

    result_leaves = leaves_list[0].copy()

    if not all(len(leaves) == len(result_leaves) for leaves in leaves_list):
        raise ValueError(
            "pytree_average received trees with differing numbers of leaves"
        )

    for idx, leaf in enumerate(result_leaves):
        if not _torch.is_tensor(leaf) or not leaf.is_floating_point():
            continue
        idx_group = [leaves[idx] for leaves in leaves_list]
        if weights is not None:
            idx_group = [weight * x for weight, x in zip(weights, idx_group)]
        stack = _torch.stack(idx_group)
        result_leaves[idx] = stack.mean(dim=0)

    result_spec = specs_list[0]
    return _pytree.tree_unflatten(result_leaves, result_spec)


def get_autocast_dtype(device_type: str) -> _torch.dtype:
    """
    Equivalent to torch.get_autocast_dtype as of torch 2.4, with fallback for
    older torch versions.
    """
    if hasattr(_torch, "get_autocast_dtype"):  # introduced in torch 2.4
        return _torch.get_autocast_dtype(device_type)

    if _torch.cuda.is_available() and device_type == "cuda":
        return _torch.get_autocast_gpu_dtype()
    if device_type == "cpu" and _torch.is_autocast_cpu_enabled():
        return _torch.get_autocast_cpu_dtype()
    return _torch.float16


def is_autocast_enabled(device_type: str) -> bool:
    """
    Equivalent to torch.is_autocast_enabled as of torch 2.4, with fallback for
    older torch versions.
    """
    from tamm._helpers import (  # pylint: disable=import-outside-toplevel
        is_torch_base_version_less_than,
    )

    # pylint: disable=too-many-function-args
    if device_type == "mps" and is_torch_base_version_less_than("2.5"):
        return False  # mps is not supported as of torch 2.4
    if hasattr(_torch, "get_autocast_dtype"):  # introduced in torch 2.4
        return _torch.is_autocast_enabled(device_type)  # device_type arg new in 2.4
    if device_type == "cpu":
        return _torch.is_autocast_cpu_enabled()
    return _torch.is_autocast_enabled()


def rms_norm(tensor, normalized_shape, weight=None, eps=None):
    if hasattr(_torch.nn.functional, "rms_norm"):  # introduced in torch 2.4
        return _torch.nn.functional.rms_norm(
            tensor, normalized_shape=normalized_shape, weight=weight, eps=eps
        )

    mean_dims = tuple(range(-len(normalized_shape), 0))
    second_moment = tensor.square().mean(dim=mean_dims, keepdim=True)
    if eps is not None:
        second_moment = second_moment + eps
    inv_norm = _torch.rsqrt(second_moment)
    if weight is None:
        return tensor * inv_norm
    return tensor * inv_norm * weight



from tamm import _helpers


class PrettyPrintListedEntities:
    def __init__(self, listing_function):
        self.listing_function = listing_function

    def print(self, include_descriptions=False, filter_deprecated=False, wide=False):
        entities = self.listing_function(
            include_descriptions=include_descriptions,
            filter_deprecated=filter_deprecated,
        )
        if not entities:
            print(
                "\n(No non-deprecated models found, check that you have permission "
                "to list them.\nUse -a to include deprecated models.)\n"
            )
            return
        if include_descriptions:
            max_len_names = max(len(name) for name in entities)
            paddings = (" " * (max_len_names + 4 - len(name)) for name in entities)
            lines = [
                name + padding + descr
                for padding, (name, descr) in zip(paddings, entities.items())
            ]
        else:
            lines = entities

        if not wide:
            lines = _helpers.truncate_lines_for_terminal(*lines)
        print("\n".join(lines))



from typing import Any as _Any
from typing import Dict as _Dict
from typing import Optional as _Optional
from typing import Union as _Union

import torch as _torch

from tamm import _helpers
from tamm.layers import side_outputs as _side_outputs
from tamm.layers.common import LayerMixin as _LayerMixin
from tamm.layers.transformer import kv_cache as _kv_cache
from tamm.layers.transformer import token_metadata as _token_metadata
from tamm.layers.transformer.layer_sequence import common as _common
from tamm.typing import ModuleOrBuilder as _ModuleOrBuilder


class KVReuseTransformerLayerSequence(
    _common.BaseTransformerLayerSequence, _LayerMixin
):
    """
    A sequence consisting of two :class:`.TransformerLayerSequence` segments.
    Rather than computing its own keys and values for self-attention, the second
    segment reuses keys and values returned from the first segment.

    The first segment must return ``keys`` and ``values`` hidden states as
    side outputs.  The second segment receives ``keys`` and ``values``
    as inputs in the ``attention_side_inputs`` argument.  The second segment
    also does not receive a ``kv_cache`` argument.

    Args:
        segment_0: A builder for the first segment.
        segment_1: A builder for the second segment.
    """

    def __init__(self, segment_0: _ModuleOrBuilder, segment_1: _ModuleOrBuilder):
        super().__init__()
        _helpers.append_children(self, segment_0=segment_0, segment_1=segment_1)

    @property
    def num_layers(self):
        return self.segment_0.num_layers + self.segment_1.num_layers

    def forward(
        self,
        hidden_states: _torch.Tensor,
        *,
        attention_side_inputs: _Dict[str, _Any],
        kv_cache: _Optional[_kv_cache.BaseKVCacheView] = None,
        secondary_attention_side_inputs: _Optional[_Dict[str, _Dict[str, _Any]]] = None,
        cross_attention_side_inputs: _Dict[str, _Any],
        token_metadata: _token_metadata.TokenMetadata,
    ) -> _Union[_torch.Tensor, _side_outputs.OutputWithSideOutputs]:
        # pylint: disable=unused-argument

        outputs_0 = self.segment_0(
            hidden_states=hidden_states,
            attention_side_inputs=attention_side_inputs,
            kv_cache=kv_cache,
            secondary_attention_side_inputs=secondary_attention_side_inputs,
            cross_attention_side_inputs=cross_attention_side_inputs,
            token_metadata=token_metadata,
        )

        side_outputs_0 = outputs_0.side_outputs

        keys, values = side_outputs_0.pop("key"), side_outputs_0.pop("value")
        if isinstance(keys, list):
            keys = keys[-1]
        if isinstance(values, list):
            values = values[-1]
        attention_side_inputs["key"] = keys
        attention_side_inputs["value"] = values

        outputs_1 = self.segment_1(
            hidden_states=outputs_0.output,
            attention_side_inputs=attention_side_inputs,
            secondary_attention_side_inputs=secondary_attention_side_inputs,
            cross_attention_side_inputs=cross_attention_side_inputs,
            token_metadata=token_metadata,
        )

        if isinstance(outputs_1, _side_outputs.OutputWithSideOutputs):
            outputs_1.side_outputs.pop("key", None)
            outputs_1.side_outputs.pop("value", None)
            outputs_1.side_outputs = _side_outputs.merge_side_outputs(
                side_outputs_0, outputs_1.side_outputs
            )
        elif len(side_outputs_0) > 0:
            outputs_1.side_outputs = _side_outputs.OutputWithSideOutputs(
                outputs_1, side_outputs=side_outputs_0
            )

        return outputs_1



# pylint: disable=no-member, cyclic-import, line-too-long

"""
_adapters_v1.model_adapters.composition
=======================================

.. autoclass:: tamm._adapters_v1.model_adapters.composition.StackedOutputTransformsModelAdapter
    :show-inheritance:
    :members:

.. autoclass:: tamm._adapters_v1.model_adapters.composition.StackedInputTransformsModelAdapter
    :show-inheritance:
    :members:

.. autoclass:: tamm._adapters_v1.model_adapters.composition.AverageOutputTransformsModelAdapter
    :show-inheritance:
    :members:

.. autoclass:: tamm._adapters_v1.model_adapters.composition.AverageInputTransformsModelAdapter
    :show-inheritance:
    :members:
"""

import logging as _logging
from typing import Dict as _Dict
from typing import Optional as _Optional

import torch.nn as _nn

from tamm._adapters_v1.adapted_layer import AdaptedLayer as _AdaptedLayer
from tamm._adapters_v1.layer_adapters import (
    AveragedInputTransforms as _AveragedInputTransforms,
)
from tamm._adapters_v1.layer_adapters import (
    AveragedOutputTransforms as _AveragedOutputTransforms,
)
from tamm._adapters_v1.layer_adapters import LayerAdapter as _LayerAdapter
from tamm._adapters_v1.layer_adapters import (
    StackedInputTransforms as _StackedInputTransforms,
)
from tamm._adapters_v1.layer_adapters import (
    StackedOutputTransforms as _StackedOutputTransforms,
)
from tamm._adapters_v1.model_adapters.model_adapter import AdapterSpec as _AdapterSpec
from tamm._adapters_v1.model_adapters.model_adapter import ModelAdapter as _ModelAdapter

_logger = _logging.getLogger(__name__)


class CompositeModelAdapterMixin:
    """
    A mixin which implements composition of multiple model adapters. Any class
    that inherits from this mixin should implement the :func:`_create_composite_adapter`
    method which consumes child :class:`LayerAdapter`s and composes them to create
    a composite adapter.
    """

    def __post_init__(self):
        # if an adapter's pretrained flag is not set, inherit it from parent
        for adapter in self.adapters.values():
            if adapter.pretrained_path is not None and adapter.pretrained is None:
                adapter.pretrained = self.pretrained

    def _create_adapters_impl(self, model: _nn.Module) -> _AdapterSpec:
        child_adapter_specs = {
            ad_id: adapter.create_adapters(model).layer_adapters
            for ad_id, adapter in self.adapters.items()
        }

        layer_adapters = {}
        for name, submodule in model.named_modules(remove_duplicate=True):
            if isinstance(submodule, _AdaptedLayer):
                child_adapters = {
                    ad_id: child_adapter_specs[ad_id][name]
                    for ad_id in self.adapters
                    if name in child_adapter_specs[ad_id]
                }
                if len(child_adapters) == 0:
                    continue
                if len(child_adapters) == 1:
                    _logger.debug(
                        f"Exactly one adapter for module {name}, so passing it through."
                    )
                    layer_adapters[name] = list(child_adapters.values())[0]
                elif len(child_adapters) < len(self.adapters):
                    no_child_adapters = set(self.adapters.keys()) - set(
                        child_adapters.keys()
                    )
                    raise ValueError(
                        f"Adapters cannot be composed for module: {name}, "
                        "because adapters with these ids do not create "
                        f"any layer adapters: {no_child_adapters}."
                    )
                else:
                    layer_adapters[name] = self._create_composite_adapter(
                        child_adapters
                    )
        return _AdapterSpec(layer_adapters=layer_adapters)


class StackedOutputTransformsModelAdapter(CompositeModelAdapterMixin, _ModelAdapter):
    """
    A composite :class:`ModelAdapter` that stacks together output transformations from
    multiple child adapters. This model adapter initializes
    :class:`StackedOutputTransforms` :class:`LayerAdapter`s for all the layers
    where layer adapters are inserted by child model adapters.

    Args:
        adapters (:obj:`dict` that maps :obj:`str` to :obj:`ModelAdapter`): A dictionary
            that maps child adapter IDs to corresponding adapters.  Note that the
            ordering of the values in ``adapters`` determines the ordering of the stack.
    """

    adapters: _Dict[str, _ModelAdapter]

    def _create_composite_adapter(self, child_adapters: _Dict[str, _LayerAdapter]):
        return _StackedOutputTransforms(child_adapters)


class StackedInputTransformsModelAdapter(CompositeModelAdapterMixin, _ModelAdapter):
    """
    A composite :class:`ModelAdapter` that stacks together input transformations from
    multiple child adapters. This model adapter initializes
    :class:`StackedInputTransforms` :class:`LayerAdapter`s for all the layers
    where layer adapters are inserted by child model adapters.

    Args:
        adapters (:obj:`dict` that maps :obj:`str` to :obj:`ModelAdapter`): A dictionary
            that maps child adapter IDs to corresponding adapters.  Note that the
            ordering of the values in ``adapters`` determines the ordering of the stack.
    """

    adapters: _Dict[str, _ModelAdapter]

    def _create_composite_adapter(self, child_adapters: _Dict[str, _LayerAdapter]):
        return _StackedInputTransforms(child_adapters)


class AveragedOutputTransformsModelAdapter(CompositeModelAdapterMixin, _ModelAdapter):
    """
    A composite :class:`ModelAdapter` that averages together output transformations from
    multiple child adapters. This model adapter initializes
    :class:`AveragedOutputTransforms` :class:`LayerAdapter`s for all the layers
    where layer adapters are inserted by child model adapters.

    Args:
        adapters (:obj:`dict` that maps :obj:`str` to :obj:`ModelAdapter`): A dictionary
            that maps child adapter IDs to corresponding model adapters.
        weights (:obj:`dict` that maps :obj:`str` to :obj:`float`, optional): A
            dictionary that maps child adapter IDs to corresponding weights for
            averaging transformed inputs.  Defaults to ``None``, which results in no
            weighting.
    """

    adapters: _Dict[str, _ModelAdapter]
    weights: _Optional[_Dict[str, float]] = None

    def _create_composite_adapter(self, child_adapters: _Dict[str, _LayerAdapter]):
        return _AveragedOutputTransforms(child_adapters, weights=self.weights)


class AveragedInputTransformsModelAdapter(CompositeModelAdapterMixin, _ModelAdapter):
    """
    A composite :class:`ModelAdapter` that averages together input transformations from
    multiple child adapters. This model adapter initializes
    :class:`AveragedInputTransforms` :class:`LayerAdapter`s for all the layers
    where layer adapters are inserted by child model adapters.

    Args:
        adapters (:obj:`dict` that maps :obj:`str` to :obj:`ModelAdapter`): A dictionary
            that maps child adapter IDs to corresponding model adapters.
        weights (:obj:`dict` that maps :obj:`str` to :obj:`float`, optional): A
            dictionary that maps child adapter IDs to corresponding weights for
            averaging transformed inputs.  Defaults to ``None``, which results in no
            weighting.
    """

    adapters: _Dict[str, _ModelAdapter]
    weights: _Optional[_Dict[str, float]] = None

    def _create_composite_adapter(self, child_adapters: _Dict[str, _LayerAdapter]):
        return _AveragedInputTransforms(child_adapters, weights=self.weights)



from typing import Any as _Any
from typing import Optional as _Optional

import torch as _torch

from tamm.layers.common import LayerMixin as _LayerMixin
from tamm.layers.mixture_of_experts.dispatcher import common as _common

# pylint: disable=line-too-long


class DroplessMixtureOfExpertsDispatcher(
    _common.MixtureOfExpertsDispatcher, _LayerMixin
):
    """
    A dispatcher that performs no token dropping.  The dispatch operation
    transforms inputs of shape ``(batch_size, seq_len, hidden_dim)`` into
    two tensors:

    1. Experts input hidden states with shape
       ``(batch_size * seq_len * num_experts_per_token, hidden_dim)``
       where the first dimension is sorted by expert index.
    2. A 1d tensor with length ``num_experts``, which contains the number of
       tokens routed to each expert.

    This dispatcher is intended for experts implemented using
    `grouped GEMM <https://developer.nvidia.com/blog/introducing-grouped-gemm-apis-in-cublas-and-more-performance-updates/#grouped_gemm_apis>`__.

    The layer also supports multiple sets of experts (each set of size
    ``num_experts``) via the ``track_dim`` argument.  For this, the layer
    maps inputs with shape ``(num_tracks, batch_size, seq_len, hidden_dim)``
    to an experts input with shape
    ``(num_tracks, batch_size * seq_len * num_experts_per_token, hidden_dim)``
    and group sizes with shape ``(num_tracks, num_experts)``.

    Args:
        num_experts (:obj:`int`): The number of experts.  If ``track_dim``
            is not ``None``, this is the number of experts per track.
        track_dim (:obj:`int`, optional): If using multiple sets (tracks) of
            experts, this is the index of the track dimension in the
            :meth:`forward` dispatching inputs.  Defaults to ``None``.
    """

    def __init__(self, num_experts: int, track_dim: _Optional[int] = None):
        super().__init__()
        self.num_experts = num_experts
        self.track_dim = track_dim

    def extra_repr(self):
        components = [f"num_experts={self.num_experts}"]
        if self.track_dim is not None:
            components.append(f"track_dim={self.track_dim}")
        return ", ".join(components)

    @property
    def _num_track_dims(self):
        return 0 if self.track_dim is None else 1

    def _dispatch(
        self, hidden_states: _torch.Tensor, *, expert_assignments: _torch.Tensor
    ) -> _common.MixtureOfExpertsDispatchOutput:
        # shorthand for shape comments:
        # - T: num tracks
        # - B: batch size
        # - S: sequence length
        # - D: hidden dim
        # - K: experts per token

        x = hidden_states

        if self.track_dim not in (0, None):
            x = x.movedim(self.track_dim, 0)  # (T, B, S, D)
            expert_assignments = expert_assignments.movedim(
                self.track_dim, 0
            )  # (T, B, S, K)

        num_experts_per_token = expert_assignments.size(-1)
        x = x.expand(num_experts_per_token, *x.shape)  # (K, T, B, S, D)
        x = x.movedim(0, -2)  # (T, B, S, K, D)
        expanded_hidden_states_shape = x.shape

        expert_assignments = expert_assignments.flatten(
            start_dim=self._num_track_dims
        )  # (T, B*S*K)
        indices = _torch.argsort(expert_assignments, dim=-1)

        x = x.flatten(start_dim=self._num_track_dims, end_dim=-2)  # (T, B*S*K, D)
        x = _torch.take_along_dim(x, indices[..., None], -2)  # (T, B*S*K, D)

        group_sizes = _torch.nn.functional.one_hot(  # pylint: disable=not-callable
            expert_assignments, num_classes=self.num_experts
        ).sum(dim=-2)

        combine_indices = _torch.argsort(indices, dim=-1)

        return _common.MixtureOfExpertsDispatchOutput(
            experts_arg=x,
            experts_kwargs={"group_sizes": group_sizes},
            combine_input={
                "combine_indices": combine_indices,
                "expanded_hidden_states_shape": expanded_hidden_states_shape,
            },
        )

    def _combine(
        self,
        experts_output: _torch.Tensor,
        *,
        assignment_weights: _torch.Tensor,
        combine_input: _Any,
    ) -> _torch.Tensor:
        combine_indices = combine_input["combine_indices"]
        expanded_hidden_states_shape = combine_input["expanded_hidden_states_shape"]

        if self.track_dim not in (0, None):
            assignment_weights = assignment_weights.movedim(
                self.track_dim, 0
            )  # (T, B, S, K)

        x = _torch.take_along_dim(
            experts_output, combine_indices[..., None], -2
        )  # (T, B*S*K, D)
        x = x.reshape(expanded_hidden_states_shape)  # (T, B, S, K, D)
        assignment_weights = assignment_weights.to(x.dtype)
        x = _torch.linalg.vecdot(  # pylint: disable=not-callable
            x, assignment_weights[..., None], dim=-2
        )  # (T, B, S, D)

        if self.track_dim not in (0, None):
            x = x.movedim(0, self.track_dim)

        return x
